"""
    Model class for page.

    Refer: https://developers.facebook.com/docs/graph-api/reference/page
"""
from dataclasses import dataclass
from typing import List, Optional

from pyfacebook.models.base import BaseModel, field
from pyfacebook.models.extensions import VOIPInfo, Paging
from pyfacebook.models.place import Location
from pyfacebook.models.image import CoverPhoto, Picture


@dataclass
class PageCategory(BaseModel):
    """
    A class representing page category.

    Refer: https://developers.facebook.com/docs/graph-api/reference/page-category/
    """

    id: Optional[str] = field(repr=True, compare=True)
    api_enum: Optional[str] = field()
    fb_page_categories: Optional[List["PageCategory"]] = field()
    name: Optional[str] = field(repr=True)


@dataclass
class MailingAddress(BaseModel):
    """
    A class representing mailing address.

    Refer: https://developers.facebook.com/docs/graph-api/reference/mailing-address/
    """

    id: Optional[str] = field(repr=True, compare=True)
    city: Optional[str] = field(repr=True)
    city_page: Optional["Page"] = field()
    country: Optional[str] = field()
    postal_code: Optional[str] = field()
    region: Optional[str] = field()
    street1: Optional[str] = field()
    street2: Optional[str] = field()


@dataclass
class Engagement(BaseModel):
    """
    A class representing the Engagement.

    Refer: https://developers.facebook.com/docs/graph-api/reference/engagement/
    """

    count: Optional[int] = field()
    count_string: Optional[str] = field()
    count_string_with_like: Optional[str] = field()
    count_string_without_like: Optional[str] = field()
    social_sentence: Optional[str] = field()
    social_sentence_with_like: Optional[str] = field()
    social_sentence_without_like: Optional[str] = field()


@dataclass
class PageParking(BaseModel):
    """
    A class representing the page parking.
    Refer: https://developers.facebook.com/docs/graph-api/reference/page-parking/
    """

    lot: Optional[int] = field(repr=True)
    street: Optional[int] = field(repr=True)
    valet: Optional[int] = field(repr=True)


@dataclass
class PageRestaurantServices(BaseModel):
    """
    A class representing the page restaurant services.

    Refer: https://developers.facebook.com/docs/graph-api/reference/page-restaurant-services/
    """

    catering: Optional[bool] = field(repr=True)
    delivery: Optional[bool] = field(repr=True)
    groups: Optional[bool] = field(repr=True)
    kids: Optional[bool] = field(repr=True)
    outdoor: Optional[bool] = field(repr=True)
    pickup: Optional[bool] = field(repr=True)
    reserve: Optional[bool] = field(repr=True)
    takeout: Optional[bool] = field(repr=True)
    waiter: Optional[bool] = field(repr=True)
    walkins: Optional[bool] = field(repr=True)


@dataclass
class PageRestaurantSpecialties(BaseModel):
    """
    A class representing the Page Restaurant Specialties.

    Refer: https://developers.facebook.com/docs/graph-api/reference/page-restaurant-specialties/
    """

    breakfast: Optional[int] = field(repr=True)
    coffee: Optional[int] = field(repr=True)
    dinner: Optional[int] = field(repr=True)
    drinks: Optional[int] = field(repr=True)
    lunch: Optional[int] = field(repr=True)


@dataclass
class PagePaymentOptions(BaseModel):
    """
    A class representing the Page Payment Options.

    Refer: https://developers.facebook.com/docs/graph-api/reference/page-payment-options/
    """

    amex: Optional[int] = field(repr=True)
    cash_only: Optional[int] = field()
    discover: Optional[int] = field()
    mastercard: Optional[int] = field()
    visa: Optional[int] = field()


@dataclass
class PageStartDate(BaseModel):
    """
    A class representing the Page Start Date.

    Refer: https://developers.facebook.com/docs/graph-api/reference/page-start-date/
    """

    day: Optional[int] = field(repr=True)
    month: Optional[int] = field(repr=True)
    year: Optional[int] = field(repr=True)


@dataclass
class PageStartInfo(BaseModel):
    """
    A class representing the Page Start Info.

    Refer: https://developers.facebook.com/docs/graph-api/reference/page-start-info/
    """

    date: Optional[PageStartDate] = field(repr=True)
    type: Optional[str] = field(repr=True)


@dataclass
class Page(BaseModel):
    """
    A class representing page.
    """

    id: Optional[str] = field(repr=True, compare=True)
    about: Optional[str] = field()
    access_token: Optional[str] = field()
    ad_campaign: Optional[dict] = field()
    affiliation: Optional[str] = field()
    app_id: Optional[str] = field()
    artists_we_like: Optional[str] = field()
    attire: Optional[str] = field()
    awards: Optional[str] = field()
    band_interests: Optional[str] = field()
    band_members: Optional[str] = field()
    best_page: Optional[dict] = field()
    bio: Optional[str] = field()
    birthday: Optional[str] = field()
    booking_agent: Optional[str] = field()
    built: Optional[str] = field()
    business: Optional[dict] = field()
    can_checkin: Optional[bool] = field()
    can_post: Optional[bool] = field()
    category: Optional[str] = field()
    category_list: Optional[List[PageCategory]] = field()
    checkins: Optional[int] = field()
    company_overview: Optional[str] = field()
    connected_instagram_account: Optional[dict] = field()
    connected_page_backed_instagram_account: Optional[dict] = field()
    contact_address: Optional[MailingAddress] = field()
    copyright_attribution_insights: Optional[dict] = field()
    copyright_whitelisted_ig_partners: Optional[List[str]] = field()
    country_page_likes: Optional[int] = field()
    cover: Optional[CoverPhoto] = field()
    culinary_team: Optional[str] = field()
    current_location: Optional[str] = field()
    delivery_and_pickup_option_info: Optional[List[str]] = field()
    description: Optional[str] = field()
    description_html: Optional[str] = field()
    differently_open_offerings: Optional[dict] = field()
    directed_by: Optional[str] = field()
    display_subtext: Optional[str] = field()
    displayed_message_response_time: Optional[str] = field()
    emails: Optional[List[str]] = field()
    engagement: Optional[Engagement] = field()
    fan_count: Optional[int] = field()
    featured_video: Optional[dict] = field()
    features: Optional[str] = field()
    followers_count: Optional[int] = field()
    food_styles: Optional[List[str]] = field()
    founded: Optional[str] = field()
    general_info: Optional[str] = field()
    general_manager: Optional[str] = field()
    genre: Optional[str] = field()
    global_brand_page_name: Optional[str] = field()
    global_brand_root_id: Optional[str] = field()
    has_added_app: Optional[bool] = field()
    has_transitioned_to_new_page_experience: Optional[bool] = field()
    has_whatsapp_business_number: Optional[bool] = field()
    has_whatsapp_number: Optional[bool] = field()
    hometown: Optional[str] = field()
    hours: Optional[dict] = field()
    impressum: Optional[str] = field()
    influences: Optional[str] = field()
    instagram_business_account: Optional[dict] = field()
    instant_articles_review_status: Optional[str] = field()
    is_always_open: Optional[bool] = field()
    is_chain: Optional[bool] = field()
    is_community_page: Optional[bool] = field()
    is_eligible_for_branded_content: Optional[bool] = field()
    is_messenger_bot_get_started_enabled: Optional[bool] = field()
    is_messenger_platform_bot: Optional[bool] = field()
    is_owned: Optional[bool] = field()
    is_permanently_closed: Optional[bool] = field()
    is_published: Optional[bool] = field()
    is_unclaimed: Optional[bool] = field()
    is_webhooks_subscribed: Optional[bool] = field()
    leadgen_tos_acceptance_time: Optional[str] = field()
    leadgen_tos_accepted: Optional[bool] = field()
    leadgen_tos_accepting_user: Optional[dict] = field()  # TODO
    link: Optional[str] = field()
    location: Optional[Location] = field()
    members: Optional[str] = field()
    merchant_id: Optional[str] = field()
    merchant_review_status: Optional[str] = field()
    messenger_ads_default_icebreakers: Optional[List[str]] = field()
    messenger_ads_default_page_welcome_message: Optional[dict] = field()
    messenger_ads_default_quick_replies: Optional[List[str]] = field()
    messenger_ads_quick_replies_type: Optional[str] = field()
    mission: Optional[str] = field()
    mpg: Optional[str] = field()
    name: Optional[str] = field(repr=True)
    name_with_location_descriptor: Optional[str] = field()
    network: Optional[str] = field()
    new_like_count: Optional[int] = field()
    offer_eligible: Optional[bool] = field()
    overall_star_rating: Optional[float] = field()
    page_token: Optional[str] = field()
    parent_page: Optional["Page"] = field()
    parking: Optional[PageParking] = field()
    payment_options: Optional[PagePaymentOptions] = field()
    personal_info: Optional[str] = field()
    personal_interests: Optional[str] = field()
    pharma_safety_info: Optional[str] = field()
    phone: Optional[str] = field()
    pickup_options: Optional[List[str]] = field()
    place_type: Optional[str] = field()
    plot_outline: Optional[str] = field()
    preferred_audience: Optional[dict] = field()
    press_contact: Optional[dict] = field()
    price_range: Optional[dict] = field()
    products: Optional[dict] = field()
    promotion_eligible: Optional[bool] = field()
    promotion_ineligible_reason: Optional[dict] = field()
    public_transit: Optional[dict] = field()
    rating_count: Optional[dict] = field()
    recipient: Optional[dict] = field()
    record_label: Optional[dict] = field()
    release_date: Optional[dict] = field()
    restaurant_services: Optional[PageRestaurantServices] = field()
    restaurant_specialties: Optional[PageRestaurantSpecialties] = field()
    tasks: Optional[List[str]] = field()
    schedule: Optional[str] = field()
    screenplay_by: Optional[str] = field()
    season: Optional[str] = field()
    single_line_address: Optional[str] = field()
    starring: Optional[str] = field()
    start_info: Optional[PageStartInfo] = field()
    store_code: Optional[str] = field()
    store_location_descriptor: Optional[str] = field()
    store_number: Optional[int] = field()
    studio: Optional[str] = field()
    supports_donate_button_in_live_video: Optional[bool] = field()
    supports_instant_articles: Optional[bool] = field()
    talking_about_count: Optional[int] = field()
    temporary_status: Optional[str] = field()
    unread_message_count: Optional[int] = field()
    unread_notif_count: Optional[int] = field()
    unseen_message_count: Optional[int] = field()
    username: Optional[str] = field()
    verification_status: Optional[str] = field()
    voip_info: Optional[VOIPInfo] = field()
    website: Optional[str] = field()
    were_here_count: Optional[int] = field()
    whatsapp_number: Optional[str] = field()
    written_by: Optional[str] = field()

    # common connection
    picture: Optional[Picture] = field()


@dataclass
class PagesSummary(BaseModel):
    """
    A class representing the aggregated information for pages
    """

    total_count: Optional[int] = field(repr=True)


@dataclass
class PagesResponse(BaseModel):
    """
    A class representing the pages response.

    Refer: https://developers.facebook.com/docs/graph-api/reference/user/accounts/
    """

    data: Optional[List[Page]] = field(repr=True)
    paging: Optional[Paging] = field()
    summary: Optional[PagesSummary] = field()


@dataclass
class SearchPage(BaseModel):
    """
    A class represent the page for search.

    Refer: https://developers.facebook.com/docs/pages/searching
    """

    id: Optional[str] = field(repr=True, compare=True)
    is_eligible_for_branded_content: Optional[bool] = field()
    is_unclaimed: Optional[bool] = field()
    link: Optional[str] = field(repr=True)
    location: Optional[Location] = field()
    name: Optional[str] = field(repr=True)
    verification_status: Optional[str] = field()


@dataclass
class SearchPagesResponse(BaseModel):
    """
    A class representing the search pages response.

    Refer: https://developers.facebook.com/docs/pages/searching
    """

    data: Optional[List[Page]] = field(repr=True)
    paging: Optional[Paging] = field()



"""
    Model class for user.

    Refer: https://developers.facebook.com/docs/graph-api/reference/user
"""

from dataclasses import dataclass
from typing import List, Optional

from dataclasses_json import config

from pyfacebook.models.base import BaseModel, field
from pyfacebook.models.image import Picture


@dataclass
class UserAgeRange(BaseModel):
    """
    A class representing user age range.

    Refer: https://developers.facebook.com/docs/graph-api/reference/age-range/
    """

    max: Optional[int] = field(repr=True)
    min: Optional[int] = field(repr=True)


@dataclass
class UserExperience(BaseModel):
    """
    A class representing user experience.

    Refer: https://developers.facebook.com/docs/graph-api/reference/experience/
    """

    id: Optional[str] = field(repr=True, compare=True)
    name: Optional[str] = field(repr=True)
    description: Optional[str] = field()
    _from: Optional[dict] = field(metadata=config(field_name="from"))
    _with: Optional[List[dict]] = field(metadata=config(field_name="with"))


@dataclass
class User(BaseModel):
    id: Optional[str] = field(repr=True, compare=True)
    about: Optional[str] = field()
    age_range: Optional[UserAgeRange] = field()
    birthday: Optional[str] = field()
    email: Optional[str] = field()
    favorite_athletes: Optional[List[UserExperience]] = field()
    favorite_teams: Optional[List[UserExperience]] = field()
    first_name: Optional[str] = field()
    gender: Optional[str] = field()
    hometown: Optional[dict] = field()
    inspirational_people: Optional[List[UserExperience]] = field()
    install_type: Optional[str] = field()
    installed: Optional[bool] = field()
    is_guest_user: Optional[bool] = field()
    languages: Optional[List[UserExperience]] = field()
    last_name: Optional[str] = field()
    link: Optional[str] = field()
    location: Optional[dict] = field()
    meeting_for: Optional[List[str]] = field()
    middle_name: Optional[str] = field()
    name: Optional[str] = field(repr=True)
    name_format: Optional[str] = field()
    payment_pricepoints: Optional[dict] = field()
    profile_pic: Optional[str] = field()
    quotes: Optional[str] = field()
    shared_login_upgrade_required_by: Optional[str] = field()
    short_name: Optional[str] = field()
    significant_other: Optional[dict] = field()
    supports_donate_button_in_live_video: Optional[bool] = field()
    video_upload_limits: Optional[dict] = field()

    # connections fields
    picture: Optional[Picture] = field()



"""
    Client for Instagram Graph API.
"""

from pyfacebook.api.base_client import BaseApi
from pyfacebook.api.instagram_business import resource as rs


class IGBusinessApi(BaseApi):
    """
    Api class for Instagram Business
    """

    user = rs.IGBusinessUser()
    media = rs.IGBusinessMedia()
    comment = rs.IGBusinessComment()
    reply = rs.IGBusinessReply()
    hashtag = rs.IGBusinessHashtag()
    container = rs.IGBusinessContainer()



"""
    Apis for hashtag.
"""

from typing import Dict, Optional, Union

import pyfacebook.utils.constant as const
from pyfacebook.api.base_resource import BaseResource
from pyfacebook.models.ig_business_models import IgBusHashtag, IgBusMediaResponse
from pyfacebook.utils.params_utils import enf_comma_separated


class IGBusinessHashtag(BaseResource):
    def get_info(
        self,
        hashtag_id: str,
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[IgBusHashtag, dict]:
        """
        Get information about a Business hashtag.

        :param hashtag_id: ID for Hashtag.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for IgBusHashtag.
            Or return json data. Default is false.
        :return: Business hashtag information.
        """
        if fields is None:
            fields = const.IG_BUSINESS_HASHTAG_PUBLIC_FIELDS

        data = self.client.get_object(
            object_id=hashtag_id,
            fields=enf_comma_separated(field="fields", value=fields),
        )
        if return_json:
            return data
        else:
            return IgBusHashtag.new_from_json_dict(data=data)

    def get_batch(
        self,
        ids: Optional[Union[str, list, tuple]],
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[Dict[str, IgBusHashtag], dict]:
        """
        Get batch business hashtags information by ids

        :param ids: IDs for the hashtags.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for IgBusHashtag.
            Or return json data. Default is false.
        :return: Business hashtags information.
        """
        ids = enf_comma_separated(field="ids", value=ids)

        if fields is None:
            fields = const.IG_BUSINESS_HASHTAG_PUBLIC_FIELDS

        data = self.client.get_objects(
            ids=ids, fields=enf_comma_separated(field="fields", value=fields)
        )
        if return_json:
            return data
        else:
            return {
                tag_id: IgBusHashtag.new_from_json_dict(item)
                for tag_id, item in data.items()
            }

    def get_top_media(
        self,
        hashtag_id: str,
        fields: Optional[Union[str, list, tuple]] = None,
        count: Optional[int] = 10,
        limit: Optional[int] = 10,
        return_json: bool = False,
    ) -> Union[IgBusMediaResponse, dict]:
        """
        Get the Most Popular Hashtagged Media

        :param hashtag_id: ID for the hashtag.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param count: The total count for you to get data.
        :param limit: Each request retrieve objects count.
            It should no more than 50. Default is None will use api default limit.
        :param return_json: Set to false will return a dataclass for IgBusMediaResponse.
            Or return json data. Default is false.
        :return: Media response information.
        """

        if fields is None:
            fields = const.IG_BUSINESS_MEDIA_PUBLIC_FIELDS

        data = self.client.get_full_connections(
            object_id=hashtag_id,
            connection="top_media",
            user_id=self.client.instagram_business_id,
            fields=enf_comma_separated(field="fields", value=fields),
            count=count,
            limit=limit,
        )

        if return_json:
            return data
        else:
            return IgBusMediaResponse.new_from_json_dict(data)

    def get_recent_media(
        self,
        hashtag_id: str,
        fields: Optional[Union[str, list, tuple]] = None,
        count: Optional[int] = 10,
        limit: Optional[int] = 10,
        return_json: bool = False,
    ) -> Union[IgBusMediaResponse, dict]:
        """
        Get the recent Hashtagged Media

        :param hashtag_id: ID for the hashtag.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param count: The total count for you to get data.
        :param limit: Each request retrieve objects count.
            It should no more than 50. Default is None will use api default limit.
        :param return_json: Set to false will return a dataclass for IgBusMediaResponse.
            Or return json data. Default is false.
        :return: Media response information.
        """

        if fields is None:
            fields = const.IG_BUSINESS_MEDIA_PUBLIC_FIELDS

        data = self.client.get_full_connections(
            object_id=hashtag_id,
            connection="recent_media",
            user_id=self.client.instagram_business_id,
            fields=enf_comma_separated(field="fields", value=fields),
            count=count,
            limit=limit,
        )

        if return_json:
            return data
        else:
            return IgBusMediaResponse.new_from_json_dict(data)



"""
    Tests for application.
"""

import responses


def test_get_info(helpers, fb_api):
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{fb_api.app_id}",
            json=helpers.load_json(
                "testdata/facebook/apidata/applications/application_info.json"
            ),
        )

        app = fb_api.application.get_info()
        assert app.id == "123456789"

        app_json = fb_api.application.get_info(
            fields="id,category,description,link,name,namespace", return_json=True
        )
        assert app_json["id"] == "123456789"


def test_get_accounts(helpers, fb_api):
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{fb_api.app_id}/accounts",
            json=helpers.load_json(
                "testdata/facebook/apidata/applications/application_accounts.json",
            ),
        )

        accounts = fb_api.application.get_accounts(count=None)
        assert len(accounts.data) == 4

        accounts_json = fb_api.application.get_accounts(
            fields="id,login_url", count=3, limit=4, return_json=True
        )
        assert accounts_json["data"][0]["id"] == "123456789"



"""
    Models for business.

    Refer: https://developers.facebook.com/docs/marketing-api/reference/business/
"""

from dataclasses import dataclass
from typing import List, Optional

from pyfacebook.models.base import BaseModel, field
from pyfacebook.models.extensions import Paging


@dataclass
class Business(BaseModel):
    """
    A class representing the Business.
    """

    id: Optional[str] = field(repr=True, compare=True)
    block_offline_analytics: Optional[bool] = field()
    collaborative_ads_managed_partner_business_info: Optional[dict] = field()
    collaborative_ads_managed_partner_eligibility: Optional[dict] = field()
    created_by: Optional[dict] = field()
    created_time: Optional[str] = field()
    extended_updated_time: Optional[str] = field()
    is_hidden: Optional[bool] = field()
    link: Optional[str] = field()
    name: Optional[str] = field(repr=True)
    payment_account_id: Optional[str] = field()
    primary_page: Optional[dict] = field()
    profile_picture_uri: Optional[str] = field()
    timezone_id: Optional[int] = field()
    two_factor_type: Optional[str] = field()
    updated_by: Optional[dict] = field()
    updated_time: Optional[str] = field()
    verification_status: Optional[str] = field()
    vertical: Optional[str] = field()
    vertical_id: Optional[int] = field()
    permitted_roles: Optional[List[str]] = field()


@dataclass
class BusinessResponse(BaseModel):
    """
    A class representing the result for businesses edge.

    Refer: https://developers.facebook.com/docs/graph-api/reference/user/businesses/
    """

    data: List[Business] = field(repr=True, compare=True)
    paging: Optional[Paging] = field()



"""
    Models for event.

    Refer: https://developers.facebook.com/docs/graph-api/reference/event
"""

from dataclasses import dataclass
from typing import List, Optional

from pyfacebook.models.base import BaseModel, field
from pyfacebook.models.image import CoverPhoto


@dataclass
class EventTime(BaseModel):
    """
    A class representing for the time of child event.
    """

    id: Optional[str] = field(repr=True, compare=True)
    start_time: Optional[str] = field(repr=True)
    ticket_uri: Optional[str] = field()
    end_time: Optional[str] = field(repr=True)


@dataclass
class Event(BaseModel):
    """
    A class representing the Event.
    """

    id: Optional[str] = field(repr=True, compare=True)
    attending_count: Optional[int] = field()
    can_guests_invite: Optional[bool] = field()
    category: Optional[str] = field()
    cover: Optional[CoverPhoto] = field()
    created_time: Optional[str] = field()
    declined_count: Optional[int] = field()
    description: Optional[str] = field()
    discount_code_enabled: Optional[bool] = field()
    end_time: Optional[str] = field()
    event_times: Optional[List[EventTime]] = field()
    guest_list_enabled: Optional[bool] = field()
    interested_count: Optional[int] = field()
    is_canceled: Optional[bool] = field()
    is_draft: Optional[bool] = field()
    is_online: Optional[bool] = field()
    is_page_owned: Optional[bool] = field()
    maybe_count: Optional[int] = field()
    name: Optional[str] = field(repr=True)
    noreply_count: Optional[int] = field()
    online_event_format: Optional[str] = field()
    online_event_third_party_url: Optional[str] = field()
    owner: Optional[dict] = field()
    parent_group: Optional[
        dict
    ] = (
        field()
    )  # TODO Refer: https://developers.facebook.com/docs/graph-api/reference/group/
    place: Optional[
        dict
    ] = (
        field()
    )  # TODO Refer: https://developers.facebook.com/docs/graph-api/reference/place/
    scheduled_publish_time: Optional[str] = field()
    start_time: Optional[str] = field(repr=True)
    ticket_uri: Optional[str] = field()
    ticket_uri_start_sales_time: Optional[str] = field()
    ticketing_privacy_uri: Optional[str] = field()
    ticketing_terms_uri: Optional[str] = field()
    timezone: Optional[str] = field()
    type: Optional[str] = field()
    updated_time: Optional[str] = field()



"""
    tests for custom exceptions
"""

from pyfacebook import LibraryError, FacebookError


def test_error():
    error = {
        "error": {
            "message": "Message describing the error",
            "type": "OAuthException",
            "code": 190,
            "error_subcode": 460,
            "error_user_title": "A title",
            "error_user_msg": "A message",
            "fbtrace_id": "EJplcsCHuLu",
        }
    }

    fb_err = FacebookError(error)

    assert fb_err.code == 190
    assert "FacebookError" in repr(fb_err)

    error = {"message": "error message"}
    lib_err = LibraryError(error)

    assert lib_err.code == -1
    assert "LibraryError" in str(lib_err)



import json

import pytest

from pyfacebook import GraphAPI


class Helpers:
    @staticmethod
    def load_json(filename):
        with open(filename, "rb") as f:
            return json.loads(f.read().decode("utf-8"))

    @staticmethod
    def load_file_binary(filename):
        with open(filename, "rb") as f:
            return f.read()


@pytest.fixture
def helpers():
    return Helpers()


@pytest.fixture
def pubg_api():
    return GraphAPI(app_id="123456", app_secret="xxxxx", access_token="token")


@pytest.fixture
def user_api():
    return GraphAPI(app_id="123456", app_secret="xxxxx", access_token="page token")



"""
    Tests for albums
"""

import responses


def test_get_info(helpers, fb_api):
    ab_id = "10153867132423553"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{ab_id}",
            json=helpers.load_json(
                "testdata/facebook/apidata/albums/album_info_fields.json"
            ),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{ab_id}",
            json=helpers.load_json(
                "testdata/facebook/apidata/albums/album_info_default.json"
            ),
        )

        album = fb_api.album.get_info(album_id=ab_id)
        assert album.id == ab_id
        assert album.updated_time == "2016-08-17T22:25:25+0000"

        album_json = fb_api.album.get_info(
            album_id=ab_id,
            fields="id,name,created_time",
            return_json=True,
        )
        assert album_json["id"] == ab_id


def test_get_batch(helpers, fb_api):
    ab_ids = ["10153867132423553", "10151067477123553"]

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}",
            json=helpers.load_json(
                "testdata/facebook/apidata/albums/albums_info_fields.json"
            ),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}",
            json=helpers.load_json(
                "testdata/facebook/apidata/albums/albums_info_default.json"
            ),
        )

        albums = fb_api.album.get_batch(ids=ab_ids)
        assert albums[ab_ids[0]].id == ab_ids[0]

        albums_json = fb_api.album.get_batch(
            ids=ab_ids,
            fields="id,name,created_time",
            return_json=True,
        )
        assert albums_json[ab_ids[1]]["id"] == ab_ids[1]



"""
    Client for facebook graph api
"""
from pyfacebook.api.base_client import BaseApi
from pyfacebook.api.facebook import resource as rs


class FacebookApi(BaseApi):
    """
    Api class for facebook
    """

    application = rs.FacebookApplication()
    business = rs.FacebookBusiness()
    user = rs.FacebookUser()
    page = rs.FacebookPage()
    post = rs.FacebookPost()
    group = rs.FacebookGroup()
    event = rs.FacebookEvent()
    photo = rs.FacebookPhoto()
    album = rs.FacebookAlbum()
    video = rs.FacebookVideo()
    live_video = rs.FacebookLiveVideo()
    comment = rs.FacebookComment()
    conversation = rs.FacebookConversation()
    message = rs.FacebookMessage()



"""
    Tests for photo api
"""
import pytest
import responses


def test_get_info(helpers, fb_api):
    photo_id = "10158249017468553"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{photo_id}",
            json=helpers.load_json(
                "testdata/facebook/apidata/photos/photo_info_fields.json"
            ),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{photo_id}",
            json=helpers.load_json(
                "testdata/facebook/apidata/photos/photo_info_default.json"
            ),
        )

        photo = fb_api.photo.get_info(photo_id=photo_id)
        assert photo.id == photo_id

        photo_json = fb_api.photo.get_info(
            photo_id=photo_id,
            fields="id,created_time",
            return_json=True,
        )
        assert photo_json["id"] == photo_id


def test_get_batch(helpers, fb_api):
    ids = ["10157415047288553", "10158249017468553"]

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}",
            json=helpers.load_json(
                "testdata/facebook/apidata/photos/photos_info_fields.json"
            ),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}",
            json=helpers.load_json(
                "testdata/facebook/apidata/photos/photos_info_default.json"
            ),
        )

        data = fb_api.photo.get_batch(ids=ids)
        assert ids[0] in data.keys()

        data_json = fb_api.photo.get_batch(
            ids=ids,
            fields="id,created_time",
            return_json=True,
        )
        assert data_json[ids[0]]["id"] == ids[0]



"""
    Tests for user api
"""

import responses


def test_get_info(helpers, fb_api):
    uid = "4"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{uid}",
            json=helpers.load_json("testdata/facebook/apidata/users/user.json"),
        )

        user = fb_api.user.get_info(user_id=uid)
        assert user.id == uid

        user_json = fb_api.user.get_info(
            user_id=uid,
            fields="id,first_name,last_name,middle_name,name,name_format,picture,short_name",
            return_json=True,
        )
        assert user_json["id"] == uid


def test_get_batches(helpers, fb_api):
    ids = ["4", "5"]

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}",
            json=helpers.load_json(
                "testdata/facebook/apidata/users/users_default.json"
            ),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}",
            json=helpers.load_json("testdata/facebook/apidata/users/users_fields.json"),
        )

        users = fb_api.user.get_batch(ids=ids, fields="")
        assert users[ids[0]].id == ids[0]

        users_json = fb_api.user.get_batch(ids=ids, return_json=True)
        assert users_json[ids[0]]["id"] == ids[0]


def test_get_feed(helpers, fb_api):
    uid = "4"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{uid}/feed",
            json=helpers.load_json(
                "testdata/facebook/apidata/posts/feed_fields_p1.json"
            ),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{uid}/feed",
            json=helpers.load_json(
                "testdata/facebook/apidata/posts/feed_fields_p2.json"
            ),
        )

        feed = fb_api.user.get_feed(object_id=uid, count=None, limit=5)
        assert len(feed.data) == 10
        assert feed.data[0].id == "4_10113477241177441"


def test_get_posts(helpers, fb_api):
    uid = "4"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{uid}/posts",
            json=helpers.load_json(
                "testdata/facebook/apidata/posts/feed_fields_p1.json"
            ),
        )

        feed_json = fb_api.user.get_posts(
            object_id=uid, count=4, limit=5, return_json=True
        )
        assert len(feed_json["data"]) == 4
        assert feed_json["data"][0]["id"] == "4_10113477241177441"


def test_get_accounts(helpers, fb_api):
    uid = "4"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{uid}/accounts",
            json=helpers.load_json(
                "testdata/facebook/apidata/users/user_accounts_p1.json"
            ),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{uid}/accounts",
            json=helpers.load_json(
                "testdata/facebook/apidata/users/user_accounts_p2.json"
            ),
        )

        accounts = fb_api.user.get_accounts(user_id=uid, count=None, limit=4)
        assert len(accounts.data) == 6
        assert accounts.data[0].access_token == "access_token"

        accounts_json = fb_api.user.get_accounts(
            user_id=uid,
            count=2,
            limit=2,
            return_json=True,
        )
        assert len(accounts_json["data"]) == 2


def test_get_businesses(helpers, fb_api):
    uid = "123"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{uid}/businesses",
            json=helpers.load_json(
                "testdata/facebook/apidata/users/user_businesses.json"
            ),
        )

        businesses = fb_api.user.get_businesses(user_id=uid, count=None, limit=2)
        assert len(businesses.data) == 2
        assert businesses.data[0].id == "123456789"

        businesses_json = fb_api.user.get_businesses(
            user_id=uid,
            count=2,
            limit=2,
            return_json=True,
        )
        assert len(businesses_json["data"]) == 2



"""
    function's to validate parameters.
"""

from typing import Optional, Union

from pyfacebook.exceptions import LibraryError


def enf_comma_separated(field: str, value: Optional[Union[str, list, tuple]]):
    """
    Check to see if field's value type belong to correct type.
    If it is, return api need value, otherwise, raise a LibraryError.
    :param field: Name of the field you want to do check.
    :param value: Values for the field.
    :return: Api needed string
    """
    try:
        # if value point with string, not check.
        if isinstance(value, str):
            value = [value]
        elif isinstance(value, (list, tuple)):
            pass
        else:
            raise LibraryError(
                {
                    "message": f"Parameter ({field}) must be single str,comma-separated str,list or tuple",
                }
            )

        # loop value list and remove repeat item
        seen = {}
        res = []
        for item in value:
            if item in seen:
                continue
            seen[item] = 1
            res.append(item)
        return ",".join(res)
    except (TypeError, ValueError):
        raise LibraryError(
            {
                "message": f"Parameter ({field}) must be single str,comma-separated str,list or tuple",
            }
        )



"""
    Tests for media.
"""

import responses


def test_get_info(helpers, api):
    media_id = "17896129349106152"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{media_id}",
            json=helpers.load_json(
                "testdata/instagram/apidata/medias/media_default.json"
            ),
        )

        media = api.media.get_info(media_id=media_id)
        assert media.id == media_id

        media_json = api.media.get_info(
            media_id=media_id,
            fields="id,caption,comments_count,children{id,media_type,media_url,permalink,shortcode,thumbnail_url,timestamp},like_count,media_type,media_url,permalink,timestamp",
            return_json=True,
        )
        assert media_json["id"] == media_id


def test_get_batch(helpers, api):
    media_ids = ["17893085207160254", "17896129349106152"]

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}",
            json=helpers.load_json(
                "testdata/instagram/apidata/medias/medias_fields.json"
            ),
        )

        medias = api.media.get_batch(ids=media_ids)
        assert medias[media_ids[0]].id == media_ids[0]

        medias_json = api.media.get_batch(
            ids=media_ids,
            fields="id,caption,comments_count,children{id,media_type,media_url,permalink,shortcode,thumbnail_url,timestamp},like_count,media_type,media_url,permalink,timestamp",
            return_json=True,
        )
        assert medias_json[media_ids[0]]["id"] == media_ids[0]


def test_get_comments(helpers, api):
    media_id = "17846368219941692"
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{media_id}/comments",
            json=helpers.load_json(
                "testdata/instagram/apidata/medias/comments_p1.json"
            ),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{media_id}/comments",
            json=helpers.load_json(
                "testdata/instagram/apidata/medias/comments_p2.json"
            ),
        )

        comments = api.media.get_comments(media_id=media_id)
        assert len(comments.data) == 4
        assert comments.data[0].id == "17858154961981086"

        comments_json = api.media.get_comments(
            media_id=media_id,
            return_json=True,
        )
        assert comments_json["data"][0]["id"] == "17892250648466172"


def test_get_children(helpers, api):
    media_id = "17846368219941692"
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{media_id}/children",
            json=helpers.load_json("testdata/instagram/apidata/medias/children.json"),
        )

        children = api.media.get_children(media_id=media_id)
        assert len(children.data) == 2
        assert children.data[0].media_type == "IMAGE"

        children_json = api.media.get_children(
            media_id=media_id,
            return_json=True,
        )
        assert children_json["data"][0]["id"] == "17924318443748703"


def test_get_insights(helpers, api):
    media_id = "17846368219941692"
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{media_id}/insights",
            json=helpers.load_json("testdata/instagram/apidata/medias/insights.json"),
        )

        insights = api.media.get_insights(
            media_id=media_id,
            metric="impressions,reach,saved",
        )
        assert len(insights.data) == 3
        assert insights.data[0].name == "impressions"

        insights_json = api.media.get_insights(
            media_id=media_id,
            metric="impressions,reach,saved",
            return_json=True,
        )
        assert insights_json["data"][0]["name"] == "impressions"


def test_get_product_tags(helpers, api):
    media_id = "90010778325754"
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{media_id}/product_tags",
            json=helpers.load_json(
                "testdata/instagram/apidata/medias/product_tags.json"
            ),
        )

        tags = api.media.get_product_tags(media_id=media_id)
        assert len(tags.data) == 1
        assert tags.data[0].product_id == 3231775643511089

        tags_json = api.media.get_product_tags(media_id=media_id, return_json=True)
        assert tags_json["data"][0]["review_status"] == "approved"



"""
    Apis for post.
"""

from typing import Dict, Optional, Union

import pyfacebook.utils.constant as const
from pyfacebook.api.base_resource import BaseResource
from pyfacebook.api.facebook.common_edges import CommentsEdge
from pyfacebook.models.post import Post
from pyfacebook.utils.params_utils import enf_comma_separated


class FacebookPost(BaseResource, CommentsEdge):
    def get_info(
        self,
        post_id: Optional[str],
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[Post, dict]:
        """
        Get information about a Facebook post

        :param post_id: ID for the post.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for post.
            Or return json data. Default is false.
        :return: Post information
        """

        if fields is None:
            fields = const.POST_PUBLIC_FIELDS + const.POST_CONNECTIONS_SUMMERY_FIELDS

        data = self.client.get_object(
            object_id=post_id, fields=enf_comma_separated(field="fields", value=fields)
        )
        if return_json:
            return data
        else:
            return Post.new_from_json_dict(data=data)

    def get_batch(
        self,
        ids: Optional[Union[str, list, tuple]],
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[Dict[str, Post], dict]:
        """
        Get batch posts information by ids.

        :param ids: IDs for the posts.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for post.
            Or return json data. Default is false.
        :return: Posts information.
        """

        ids = enf_comma_separated(field="ids", value=ids)

        if fields is None:
            fields = const.POST_PUBLIC_FIELDS + const.POST_CONNECTIONS_SUMMERY_FIELDS

        data = self.client.get_objects(
            ids=ids, fields=enf_comma_separated(field="fields", value=fields)
        )
        if return_json:
            return data
        else:
            return {
                post_id: Post.new_from_json_dict(item) for post_id, item in data.items()
            }



"""
    Model for video.

    Refer: https://developers.facebook.com/docs/graph-api/reference/video/
"""

from dataclasses import dataclass
from typing import List, Optional

from dataclasses_json import config

from pyfacebook.models.base import BaseModel, field
from pyfacebook.models.extensions import Paging, Privacy


@dataclass
class VideoFormat(BaseModel):
    """
    A class representing the Video Format.

    Refer: https://developers.facebook.com/docs/graph-api/reference/video-format/
    """

    embed_html: Optional[str] = field()
    filter: Optional[str] = field(repr=True)
    height: Optional[int] = field()
    picture: Optional[str] = field()
    width: Optional[int] = field()


@dataclass
class VideoStatus(BaseModel):
    """
    A class representing the Video Status.

    Refer: https://developers.facebook.com/docs/graph-api/reference/video-status/
    """

    processing_progress: Optional[int] = field(repr=True)
    video_status: Optional[str] = field(repr=True)


@dataclass
class Video(BaseModel):
    """
    A class representing the Video.
    """

    id: Optional[str] = field(repr=True, compare=True)
    ad_breaks: Optional[List[int]] = field()
    backdated_time: Optional[str] = field()
    backdated_time_granularity: Optional[str] = field()
    content_category: Optional[str] = field()
    content_tags: Optional[List[int]] = field()
    created_time: Optional[str] = field()
    custom_labels: Optional[List[str]] = field()
    description: Optional[str] = field(repr=True)
    embed_html: Optional[str] = field()
    embeddable: Optional[bool] = field()
    event: Optional[
        dict
    ] = (
        field()
    )  # TODO Refer: https://developers.facebook.com/docs/graph-api/reference/event/
    format: Optional[List[VideoFormat]] = field()
    _from: Optional[dict] = field(metadata=config(field_name="from"))
    icon: Optional[str] = field()
    is_crosspost_video: Optional[bool] = field()
    is_crossposting_eligible: Optional[bool] = field()
    is_episode: Optional[bool] = field()
    is_instagram_eligible: Optional[bool] = field()
    is_reference_only: Optional[bool] = field()
    length: Optional[float] = field()
    live_status: Optional[str] = field()
    music_video_copyright: Optional[dict] = field()  # TODO
    permalink_url: Optional[str] = field()
    place: Optional[
        dict
    ] = (
        field()
    )  # TODO Refer: https://developers.facebook.com/docs/graph-api/reference/place/
    post_views: Optional[int] = field()
    premiere_living_room_status: Optional[str] = field()
    privacy: Optional[Privacy] = field()
    published: Optional[bool] = field()
    scheduled_publish_time: Optional[str] = field()
    source: Optional[str] = field()
    status: Optional[VideoStatus] = field()
    title: Optional[str] = field()
    universal_video_id: Optional[str] = field()
    updated_time: Optional[str] = field()
    views: Optional[int] = field()


@dataclass
class VideosResponse(BaseModel):
    """
    A class represent the result for videos edge.

    Refer: https://developers.facebook.com/docs/graph-api/reference/page/videos
    """

    data: List[Video] = field(repr=True, compare=True)
    paging: Optional[Paging] = field()



"""
    Models for comment.

    Refer: https://developers.facebook.com/docs/graph-api/reference/comment
"""

from dataclasses import dataclass
from typing import List, Optional

from dataclasses_json import config

from pyfacebook.models.base import BaseModel, field
from pyfacebook.models.extensions import Paging
from pyfacebook.models.attachment import StoryAttachment


@dataclass
class MessageTag(BaseModel):
    """
    A class representing the Message Tag.
    """

    id: Optional[str] = field(repr=True, compare=True)
    name: Optional[str] = field(repr=True)
    type: Optional[str] = field()
    offset: Optional[int] = field()
    length: Optional[int] = field()


@dataclass
class Comment(BaseModel):
    """
    A class representing the Comment.
    """

    id: Optional[str] = field(repr=True, compare=True)
    attachment: Optional[StoryAttachment] = field()
    can_comment: Optional[bool] = field()
    can_remove: Optional[bool] = field()
    can_hide: Optional[bool] = field()
    can_like: Optional[bool] = field()
    can_reply_privately: Optional[bool] = field()
    comment_count: Optional[int] = field()
    created_time: Optional[str] = field()
    _from: Optional[dict] = field(metadata=config(field_name="from"))
    like_count: Optional[int] = field()
    message: Optional[str] = field()
    message_tags: Optional[List[MessageTag]] = field()
    object: Optional[dict] = field()  # TODO
    parent: Optional["Comment"] = field()
    permalink_url: Optional[str] = field()
    private_reply_conversation: Optional[
        dict
    ] = (
        field()
    )  # TODO Refer: https://developers.facebook.com/docs/graph-api/reference/conversation
    user_likes: Optional[bool] = field()


@dataclass
class CommentsSummary(BaseModel):
    """
    A class representing the summary for comments.

    Refer: https://developers.facebook.com/docs/graph-api/reference/object/comments
    """

    order: Optional[str] = field(repr=True)
    total_count: Optional[int] = field(repr=True)
    can_comment: Optional[bool] = field()


@dataclass
class CommentsResponse(BaseModel):
    """
    A class representing the result for comments edge.

    Refer: https://developers.facebook.com/docs/graph-api/reference/object/comments
    """

    data: List[Comment] = field(repr=True, compare=True)
    paging: Optional[Paging] = field()
    summary: Optional[CommentsSummary] = field(repr=True)



"""
    Albums edge for resource.
"""

from typing import Optional, Union

import pyfacebook.utils.constant as const
from pyfacebook.models.album import AlbumResponse
from pyfacebook.utils.params_utils import enf_comma_separated


class AlbumsEdge:
    __slots__ = ()

    def get_albums(
        self,
        object_id: str,
        fields: Optional[Union[str, list, dict]] = None,
        since: Optional[str] = None,
        until: Optional[str] = None,
        count: Optional[int] = 10,
        limit: Optional[int] = 10,
        return_json: bool = False,
        **kwargs,
    ) -> Union[AlbumResponse, dict]:
        """
        Get a list of Albums on a Facebook object.

        :param object_id: ID for object(like page,group,user..).
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param since: A Unix timestamp or strtotime data value that points to the start of data.
        :param until: A Unix timestamp or strtotime data value that points to the end of data.
        :param count: The total count for you to get data.
        :param limit: Each request retrieve objects count.
            It should no more than 100. Default is None will use api default limit.
        :param return_json: Set to false will return a dataclass for Album.
            Or return json data. Default is false.
        :param kwargs: Additional parameters for different object.
        :return: Albums response information
        """

        if fields is None:
            fields = const.ALBUM_PUBLIC_FIELDS

        data = self.client.get_full_connections(
            object_id=object_id,
            connection="albums",
            count=count,
            limit=limit,
            fields=enf_comma_separated(field="fields", value=fields),
            since=since,
            until=until,
            **kwargs,
        )

        if return_json:
            return data
        else:
            return AlbumResponse.new_from_json_dict(data)



"""
    tests for ig basic display api
"""
import pytest
import responses

from pyfacebook import BasicDisplayAPI
from pyfacebook.exceptions import LibraryError


def test_oath_flow(helpers):
    api = BasicDisplayAPI(app_id="id", app_secret="secret", oauth_flow=True)

    # test get authorization url
    _, state = api.get_authorization_url()

    # if user give authorize.
    resp = "https://localhost/?code=code&state=PyFacebook#_"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.POST,
            url=api.EXCHANGE_ACCESS_TOKEN_URL,
            json=helpers.load_json("testdata/base/basic_display_api_user_token.json"),
        )

        r = api.exchange_user_access_token(response=resp)
        assert r["access_token"] == "token"


def test_exchange_long_lived_token(helpers):
    api = BasicDisplayAPI(access_token="token")

    # test exchange long-lived page token
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.instagram.com/access_token",
            json=helpers.load_json("testdata/base/long_term_token.json"),
        )

        res = api.exchange_long_lived_user_access_token()
        assert res["access_token"] == "token"


def test_refresh_token(helpers):
    api = BasicDisplayAPI(access_token="token")

    # test exchange long-lived page token
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.instagram.com/refresh_access_token",
            json=helpers.load_json("testdata/base/long_term_token.json"),
        )

        res = api.refresh_access_token(access_token=api.access_token)
        assert res["access_token"] == "token"


def test_not_implemented_methods():
    api = BasicDisplayAPI(access_token="token")
    with pytest.raises(LibraryError):
        api.exchange_page_access_token(page_id="id")

    with pytest.raises(LibraryError):
        api.exchange_long_lived_page_access_token(user_id="id")

    with pytest.raises(LibraryError):
        api.get_app_token()

    with pytest.raises(LibraryError):
        api.debug_token(input_token="token")



"""
    Models for common
"""

from dataclasses import dataclass
from typing import List, Optional

from pyfacebook.models.base import BaseModel, field


@dataclass
class VOIPInfo(BaseModel):
    """
    A class representing the VOIP Info.

    Refer: https://developers.facebook.com/docs/graph-api/reference/voip-info/
    """

    has_mobile_app: Optional[bool] = field(repr=True)
    has_permission: Optional[bool] = field(repr=True)
    is_callable: Optional[bool] = field(repr=True)
    is_callable_webrtc: Optional[bool] = field(repr=True)
    is_pushable: Optional[bool] = field(repr=True)
    reason_code: Optional[int] = field(repr=True)
    reason_description: Optional[str] = field(repr=True)


@dataclass
class PagingCursors(BaseModel):
    """
    A class representing the paging cursors.
    """

    after: Optional[str] = field(repr=True)
    before: Optional[str] = field()


@dataclass
class Paging(BaseModel):
    """

    Refer: https://developers.facebook.com/docs/graph-api/using-graph-api/#paging
    """

    cursors: Optional[PagingCursors] = field()
    previous: Optional[str] = field()
    next: Optional[str] = field(repr=True)


@dataclass
class Privacy(BaseModel):
    """
    A class representing the Privacy.

    Refer: https://developers.facebook.com/docs/graph-api/reference/privacy/
    """

    allow: Optional[str] = field()
    deny: Optional[str] = field()
    description: Optional[str] = field()
    friends: Optional[str] = field()
    networks: Optional[str] = field()
    value: Optional[str] = field(repr=True)


@dataclass
class ReactionSummary(BaseModel):
    """
    A class representing
    """

    total_count: Optional[int] = field()
    viewer_reaction: Optional[str] = field()


@dataclass
class Reactions(BaseModel):
    """
    A class representing the Reactions connection.

    Refer: https://developers.facebook.com/docs/graph-api/reference/post/reactions/
    """

    data: Optional[List[dict]] = field()
    paging: Optional[Paging] = field()
    summary: Optional[ReactionSummary] = field(repr=True)



"""
    Models for Attachments

    Refer: https://developers.facebook.com/docs/graph-api/reference/post/attachments/
"""

from dataclasses import dataclass
from typing import List, Optional

from pyfacebook.models.base import BaseModel, field
from pyfacebook.models.extensions import Paging
from pyfacebook.models.image import ImageSource


@dataclass
class EntityAtTextRange(BaseModel):
    """
    A class representing the Entity At Text Range.

    Refer: https://developers.facebook.com/docs/graph-api/reference/entity-at-text-range/
    """

    id: Optional[str] = field(repr=True, compare=True)
    length: Optional[int] = field()
    name: Optional[str] = field(repr=True)
    object: Optional[dict] = field()  # TODO
    offset: Optional[int] = field()
    type: Optional[str] = field()


@dataclass
class StoryAttachmentMedia(BaseModel):
    """
    A class representing the Story Attachment Media.

    Refer: https://developers.facebook.com/docs/graph-api/reference/story-attachment-media/
    """

    image: Optional[ImageSource] = field()
    source: Optional[str] = field(repr=True)


@dataclass
class StoryAttachmentTarget(BaseModel):
    """
    A class representing the Story Attachment Target.

    Refer: https://developers.facebook.com/docs/graph-api/reference/story-attachment-target/
    """

    id: Optional[str] = field(repr=True)
    unshimmed_url: Optional[str] = field()
    url: Optional[str] = field()


@dataclass
class StoryAttachmentSubattachments(BaseModel):
    """
    A class representing the Story Attachment Subattachments.

    Refer: https://developers.facebook.com/docs/graph-api/reference/story-attachment/subattachments/
    """

    data: Optional[List["StoryAttachment"]] = field(repr=True)
    paging: Optional[Paging] = field()


@dataclass
class StoryAttachment(BaseModel):
    """
    A class representing the Story Attachment

    Refer: https://developers.facebook.com/docs/graph-api/reference/story-attachment/
    """

    description: Optional[str] = field()
    description_tags: Optional[List[EntityAtTextRange]] = field()
    media: Optional[StoryAttachmentMedia] = field()
    media_type: Optional[str] = field(repr=True)
    target: Optional[StoryAttachmentTarget] = field()
    title: Optional[str] = field()
    type: Optional[str] = field(repr=True)
    unshimmed_url: Optional[str] = field()
    url: Optional[str] = field()

    # common connections
    subattachments: Optional[StoryAttachmentSubattachments] = field()


@dataclass
class Attachments(BaseModel):
    """
    A class representing the Attachments connection.

    Refer: https://developers.facebook.com/docs/graph-api/reference/post/attachments/
    """

    data: Optional[List[StoryAttachment]] = field(repr=True)
    paging: Optional[Paging] = field()



"""
    Apis for application.
"""

from typing import Dict, Optional, Union

import pyfacebook.utils.constant as const
from pyfacebook.api.base_resource import BaseResource
from pyfacebook.models.application import Application, ApplicationAccountsResponse
from pyfacebook.utils.params_utils import enf_comma_separated


class FacebookApplication(BaseResource):
    def get_info(
        self,
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[Application, dict]:
        """
        Get information for current facebook Application. Need app token.

        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for Application.
            Or return json data. Default is false.
        :return: Application information.
        """
        if fields is None:
            fields = const.APPLICATION_PUBLIC_FIELDS

        data = self.client.get_object(
            object_id=self.client.app_id,
            fields=enf_comma_separated(field="fields", value=fields),
        )
        if return_json:
            return data
        else:
            return Application.new_from_json_dict(data=data)

    def get_accounts(
        self,
        fields: Optional[Union[str, list, dict]] = None,
        count: Optional[int] = 10,
        limit: Optional[int] = 10,
        return_json: bool = False,
    ) -> Union[ApplicationAccountsResponse, dict]:
        """
        Represents a collection of test users on an app.

        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param count: The total count for you to get data.
        :param limit: Each request retrieve objects count.
            It should no more than 100. Default is None will use api default limit.
        :param return_json: Set to false will return a dataclass for post.
            Or return json data. Default is false.
        :return: Application accounts information.
        """

        if fields is None:
            fields = const.APPLICATION_ACCOUNT_PUBLIC_FIELDS

        data = self.client.get_full_connections(
            object_id=self.client.app_id,
            connection="accounts",
            count=count,
            limit=limit,
            fields=enf_comma_separated(field="fields", value=fields),
        )
        if return_json:
            return data
        else:
            return ApplicationAccountsResponse.new_from_json_dict(data)



"""
    Resource base class
"""
from pyfacebook.api.graph import GraphAPI


class BaseResource:
    """Resource base class"""

    def __init__(self, client=None):
        self._client: GraphAPI = client

    @property
    def access_token(self):
        return self._client.access_token

    @property
    def app_id(self):
        return self._client.app_id

    @property
    def app_secret(self):
        return self._client.app_secret

    @property
    def client(self):
        return self._client



"""
    Apis for user.
"""

from typing import Optional, Union

import pyfacebook.utils.constant as const
from pyfacebook.api.base_resource import BaseResource
from pyfacebook.models.ig_business_models import (
    IgBusUser,
    IgBusMediaResponse,
    IgBusDiscoveryUserResponse,
    IgBusDiscoveryUserMediaResponse,
    IgBusPublishLimitResponse,
    IgBusInsightsResponse,
    IgBusMentionedCommentResponse,
    IgBusMentionedMediaResponse,
    IgBusHashtagsResponse,
    IgBusCatalogsResponse,
    IgBusProductsResponse,
    IgBusProductAppealsResponse,
)
from pyfacebook.utils.params_utils import enf_comma_separated


class IGBusinessUser(BaseResource):
    def get_info(
        self,
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[IgBusUser, dict]:
        """
        Get information about an Instagram Business Account or an Instagram Creator Account.

        Note:

            Now this can only get user info by user's access token.

        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for Account.
            Or return json data. Default is false.
        :return: IG Business User information.
        """
        if fields is None:
            fields = const.IG_BUSINESS_USER_PUBLIC_FIELDS

        data = self.client.get_object(
            object_id=self.client.instagram_business_id,
            fields=enf_comma_separated(field="fields", value=fields),
        )
        if return_json:
            return data
        else:
            return IgBusUser.new_from_json_dict(data=data)

    def discovery_user(
        self,
        username: str,
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[IgBusDiscoveryUserResponse, dict]:
        """
        Get other business account info by username.

        :param username: Username to get data.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for IgBusDiscoveryUserResponse.
            Or return json data. Default is false.
        :return: Business discovery user response information.
        """

        if fields is None:
            fields = const.IG_BUSINESS_USER_PUBLIC_FIELDS

        metric = enf_comma_separated(field="fields", value=fields)

        data = self.client.get_object(
            object_id=self.client.instagram_business_id,
            fields=f"business_discovery.username({username}){{{metric}}}",
        )
        if return_json:
            return data
        else:
            return IgBusDiscoveryUserResponse.new_from_json_dict(data=data)

    def discovery_user_medias(
        self,
        username: str,
        fields: Optional[Union[str, list, tuple]] = None,
        limit: Optional[int] = None,
        since: Optional[str] = None,
        until: Optional[str] = None,
        after: Optional[str] = None,
        before: Optional[str] = None,
        return_json: bool = False,
    ) -> Union[IgBusDiscoveryUserMediaResponse, dict]:
        """
        Get other business account's media by username.

        :param username: Username to get data.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param limit: Each request retrieve objects count.
            It should no more than 100. Default is None will use api default limit.
        :param since: A Unix timestamp or strtotime data value that points to the start of data.
        :param until: A Unix timestamp or strtotime data value that points to the end of data.
        :param after: The cursor that points to the end of the page of data that has been returned.
        :param before: The cursor that points to the start of the page of data that has been returned.
        :param return_json: Set to false will return a dataclass for IgBusDiscoveryUserMediaResponse.
            Or return json data. Default is false.
        :return: Business discovery media response information.
        """

        if fields is None:
            fields = const.IG_BUSINESS_MEDIA_PUBLIC_FIELDS
        metric = enf_comma_separated(field="fields", value=fields)

        limit_str = f".limit({limit})" if limit is not None else ""
        after_str = f".after({after})" if after is not None else ""
        before_str = f".before({before})" if before is not None else ""
        since_str = f".since({since})" if since is not None else ""
        until_str = f".until({until})" if until is not None else ""

        data = self.client.get_object(
            object_id=self.client.instagram_business_id,
            fields=(
                f"business_discovery.username({username})"
                f"{{media{after_str}{before_str}{since_str}{until_str}{limit_str}{{{metric}}}}}"
            ),
        )

        if return_json:
            return data
        else:
            return IgBusDiscoveryUserMediaResponse.new_from_json_dict(data)

    def get_content_publishing_limit(
        self,
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[IgBusPublishLimitResponse, dict]:
        """
        Get user's current content publishing usage.

        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for IgBusPublishLimitResponse.
            Or return json data. Default is false.
        :return: Business content publish limit response information.
        """

        if fields is None:
            fields = const.IG_BUSINESS_CONTENT_PUBLISH_LIMIT_FIELDS

        data = self.client.get_connection(
            object_id=self.client.instagram_business_id,
            connection="content_publishing_limit",
            fields=enf_comma_separated(field="fields", value=fields),
        )
        if return_json:
            return data
        else:
            return IgBusPublishLimitResponse.new_from_json_dict(data)

    def get_insights(
        self,
        metric: Union[str, list, tuple],
        period: str,
        since: Optional[str] = None,
        until: Optional[str] = None,
        user_id: Optional[str] = None,
        access_token: Optional[str] = None,
        return_json: bool = False,
    ) -> Union[IgBusInsightsResponse, dict]:
        """
        Get social interaction metrics on an IG User.

        :param metric: Comma-separated id string for insights metrics which you want.
            You can also pass this with an id list, tuple.
        :param period: Period to aggregation data.
            Accepted parameters: lifetime,day,week,days_28
        :param since: Lower bound of the time range to fetch data. Need Unix timestamps.
        :param until: Upper bound of the time range to fetch data. Need Unix timestamps.
            Notice: time range may not more than 30 days.
        :param user_id: ID for business user to get insights.
        :param access_token: Access token with permissions for user_id.
        :param return_json: Set to false will return a dataclass for Insights.
            Or return json data. Default is false.
        :return: Business user insights response information.
        """

        if user_id is None:
            user_id = self.client.instagram_business_id

        args = {
            "metric": enf_comma_separated(field="metric", value=metric),
            "period": period,
            "since": since,
            "until": until,
        }
        if access_token:
            args["access_token"] = access_token

        data = self.client.get_connection(
            object_id=user_id,
            connection="insights",
            **args,
        )
        if return_json:
            return data
        else:
            return IgBusInsightsResponse.new_from_json_dict(data)

    def get_media(
        self,
        fields: Optional[Union[str, list, tuple]] = None,
        since: Optional[str] = None,
        until: Optional[str] = None,
        count: Optional[int] = 10,
        limit: Optional[int] = 10,
        return_json: bool = False,
    ) -> Union[IgBusMediaResponse, dict]:
        """
        Get user's media. This only can return max 10k medias.

        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param since: A Unix timestamp or strtotime data value that points to the start of data.
        :param until: A Unix timestamp or strtotime data value that points to the end of data.
        :param count: The total count for you to get data.
        :param limit: Each request retrieve objects count.
            It should no more than 100. Default is None will use api default limit.
        :param return_json: Set to false will return a dataclass for IgBusMediaResponse.
            Or return json data. Default is false.
        :return: Media response information.
        """

        if fields is None:
            fields = const.IG_BUSINESS_MEDIA_PUBLIC_FIELDS

        data = self.client.get_full_connections(
            object_id=self.client.instagram_business_id,
            connection="media",
            count=count,
            limit=limit,
            fields=enf_comma_separated(field="fields", value=fields),
            since=since,
            until=until,
        )
        if return_json:
            return data
        else:
            return IgBusMediaResponse.new_from_json_dict(data)

    def get_live_media(
        self,
        fields: Optional[Union[str, list, tuple]] = None,
        since: Optional[str] = None,
        until: Optional[str] = None,
        count: Optional[int] = 10,
        limit: Optional[int] = 10,
        return_json: bool = False,
    ) -> Union[IgBusMediaResponse, dict]:
        """
        Represents a collection of live video IG Media on an IG User.
        Only live video IG Media being broadcast at the time of the request will be returned.

        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param since: A Unix timestamp or strtotime data value that points to the start of data.
        :param until: A Unix timestamp or strtotime data value that points to the end of data.
        :param count: The total count for you to get data.
        :param limit: Each request retrieve objects count.
            It should no more than 100. Default is None will use api default limit.
        :param return_json: Set to false will return a dataclass for IgBusMediaResponse.
            Or return json data. Default is false.
        :return: Media response information.
        """

        if fields is None:
            fields = const.IG_BUSINESS_MEDIA_PUBLIC_FIELDS

        data = self.client.get_full_connections(
            object_id=self.client.instagram_business_id,
            connection="live_media",
            count=count,
            limit=limit,
            fields=enf_comma_separated(field="fields", value=fields),
            since=since,
            until=until,
        )
        if return_json:
            return data
        else:
            return IgBusMediaResponse.new_from_json_dict(data)

    def get_mentioned_comment(
        self,
        comment_id: str,
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[IgBusMentionedCommentResponse, dict]:
        """
        Get data on an IG Comment in which user has been @mentioned by another Instagram user

        :param comment_id: ID for comment which the user has been @mentioned.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for IgBusMentionedCommentResponse.
            Or return json data. Default is false.
        :return: Mentioned comment response.
        """

        if fields is None:
            fields = const.IG_BUSINESS_COMMENT_PUBLIC_FIELDS
        fields_inline = enf_comma_separated(field="fields", value=fields)

        data = self.client.get_object(
            object_id=self.client.instagram_business_id,
            fields=f"mentioned_comment.comment_id({comment_id}){{{fields_inline}}}",
        )

        if return_json:
            return data
        else:
            return IgBusMentionedCommentResponse.new_from_json_dict(data)

    def get_mentioned_media(
        self,
        media_id: str,
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[IgBusMentionedMediaResponse, dict]:
        """
        Get data on an IG Media in which user has been @mentioned in a caption by another Instagram user.

        :param media_id: ID for media which the user has been @mentioned in a caption
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for IgBusMentionedMediaResponse.
            Or return json data. Default is false.
        :return: Mentioned media response.
        """

        if fields is None:
            fields = const.IG_BUSINESS_MENTION_MEDIA_FIELDS
        fields_inline = enf_comma_separated(field="fields", value=fields)

        data = self.client.get_object(
            object_id=self.client.instagram_business_id,
            fields=f"mentioned_media.media_id({media_id}){{{fields_inline}}}",
        )
        if return_json:
            return data
        else:
            return IgBusMentionedMediaResponse.new_from_json_dict(data)

    def get_hashtag_search(
        self,
        q: str,
        return_json: bool = False,
    ) -> Union[IgBusHashtagsResponse, dict]:
        """
        Get ID for hashtag.

        :param q: The hashtag name to query.
        :param return_json: Set to false will return a dataclass for IgBusHashtagsResponse.
            Or return json data. Default is false.
        :return: User stories response information.
        """
        data = self.client.get(
            path="ig_hashtag_search",
            args={
                "fields": "id,name",
                "user_id": self.client.instagram_business_id,
                "q": q,
            },
        )
        if return_json:
            return data
        else:
            return IgBusHashtagsResponse.new_from_json_dict(data)

    def get_recently_searched_hashtags(
        self,
        fields: Optional[Union[str, list, tuple]] = None,
        count: Optional[int] = 25,
        limit: Optional[int] = 25,
        return_json: bool = False,
    ) -> Union[IgBusHashtagsResponse, dict]:
        """
        Get the IG Hashtags that user has searched for within the last 7 days.

        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param count: The total count for you to get data. Maximum is 30.
        :param limit: Each request retrieve objects count. Up to 30.
        :param return_json: Set to false will return a dataclass for IgBusHashtagsResponse.
            Or return json data. Default is false.
        :return: Recently searched hashtags response information.
        """

        if fields is None:
            fields = const.IG_BUSINESS_HASHTAG_PUBLIC_FIELDS

        data = self.client.get_full_connections(
            object_id=self.client.instagram_business_id,
            connection="recently_searched_hashtags",
            fields=enf_comma_separated(field="fields", value=fields),
            count=count,
            limit=limit,
        )
        if return_json:
            return data
        else:
            return IgBusHashtagsResponse.new_from_json_dict(data)

    def get_stories(
        self,
        fields: Optional[Union[str, list, tuple]] = None,
        count: Optional[int] = 10,
        limit: Optional[int] = 10,
        return_json: bool = False,
    ) -> Union[IgBusMediaResponse, dict]:
        """
        Get list of story IG Media objects on user.

        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param count: The total count for you to get data.
        :param limit: Each request retrieve objects count.
            It should no more than 100. Default is None will use api default limit.
        :param return_json: Set to false will return a dataclass for IgBusMediaResponse.
            Or return json data. Default is false.
        :return: User stories response information.
        """

        if fields is None:
            fields = const.IG_BUSINESS_MEDIA_PUBLIC_FIELDS

        data = self.client.get_full_connections(
            object_id=self.client.instagram_business_id,
            connection="stories",
            count=count,
            limit=limit,
            fields=enf_comma_separated(field="fields", value=fields),
        )
        if return_json:
            return data
        else:
            return IgBusMediaResponse.new_from_json_dict(data)

    def get_tagged_media(
        self,
        fields: Optional[Union[str, list, tuple]] = None,
        count: Optional[int] = 25,
        limit: Optional[int] = 25,
        return_json: bool = False,
    ) -> Union[IgBusMediaResponse, dict]:
        """
        Get list of Media objects in which user has been tagged by another Instagram user.

        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param count: The total count for you to get data.
        :param limit: Each request retrieve objects count.
            It should no more than 100. Default is None will use api default limit.
        :param return_json: Set to false will return a dataclass for IgBusMediaResponse.
            Or return json data. Default is false.
        :return: Tagged user medias response information.
        """

        if fields is None:
            fields = const.IG_BUSINESS_MEDIA_PUBLIC_FIELDS

        data = self.client.get_full_connections(
            object_id=self.client.instagram_business_id,
            connection="tags",
            count=count,
            limit=limit,
            fields=enf_comma_separated(field="fields", value=fields),
        )
        if return_json:
            return data
        else:
            return IgBusMediaResponse.new_from_json_dict(data)

    def get_available_catalogs(
        self, return_json: bool = False
    ) -> Union[IgBusCatalogsResponse, dict]:
        """
        Get the product catalog in an IG User's Instagram Shop.

        :param: return_json:  Set to false will return a dataclass for IgBusCatalogsResponse.
            Or return json data. Default is false.
        :return: catalog data
        """

        data = self.client.get_connection(
            object_id=self.client.instagram_business_id,
            connection="available_catalogs",
        )
        if return_json:
            return data
        else:
            return IgBusCatalogsResponse.new_from_json_dict(data)

    def get_catalog_product_search(
        self,
        catalog_id: str,
        q: Optional[str] = None,
        count: Optional[int] = 10,
        limit: int = 10,
        return_json: bool = False,
    ) -> Union[IgBusProductsResponse, dict]:
        """
        Get a collection of products that match a given search string within the targeted IG User's Instagram Shop catalog.

        :param catalog_id: ID of catalog to search.
        :param q: A string to search for in each product's name or SKU number (SKU numbers can be added in the Content ID column in the catalog management interface).
            If no string is specified, all tag-eligible products will be returned.
        :param count: The count will retrieve objects. Default is None will get all data.
        :param limit: Each request retrieve objects count.
            For most connections should no more than 100. Default is None will use api default limit.
        :param return_json: Set to false will return a dataclass for IgBusProductsResponse.
            Or return json data. Default is false.
        :return: Catalog products data.
        """

        data = self.client.get_full_connections(
            object_id=self.client.instagram_business_id,
            connection="catalog_product_search",
            catalog_id=catalog_id,
            q=q,
            count=count,
            limit=limit,
        )
        if return_json:
            return data
        else:
            return IgBusProductsResponse.new_from_json_dict(data)

    def get_product_appeal(
        self,
        product_id: str,
        return_json: bool = False,
    ) -> Union[IgBusProductAppealsResponse, dict]:
        """
        Get appeal status of a rejected product.

        :param product_id: Product ID.
        :param return_json: Set to false will return a dataclass for IgBusProductAppealsResponse.
            Or return json data. Default is false.
        :return: Product appeals data.
        """

        data = self.client.get_connection(
            object_id=self.client.instagram_business_id,
            connection="product_appeal",
            product_id=product_id,
        )
        if return_json:
            return data
        else:
            return IgBusProductAppealsResponse.new_from_json_dict(data)



"""
    Apis for event.
"""

from typing import Dict, Optional, Union

import pyfacebook.utils.constant as const
from pyfacebook.api.base_resource import BaseResource
from pyfacebook.api.facebook.common_edges import PhotosEdge, LiveVideosEdge
from pyfacebook.models.event import Event
from pyfacebook.utils.params_utils import enf_comma_separated


class FacebookEvent(BaseResource, PhotosEdge, LiveVideosEdge):
    def get_info(
        self,
        event_id: Optional[str],
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[Event, dict]:
        """
        Get information about a Facebook Event.

        :param event_id: ID for the Event.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for Event.
            Or return json data. Default is false.
        :return: Event information.
        """

        if fields is None:
            fields = const.EVENT_PUBLIC_FIELDS

        data = self.client.get_object(
            object_id=event_id,
            fields=enf_comma_separated(field="fields", value=fields),
        )
        if return_json:
            return data
        else:
            return Event.new_from_json_dict(data=data)

    def get_batch(
        self,
        ids: Optional[Union[str, list, tuple]],
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[Dict[str, Event], dict]:
        """
        Get batch Events information by ids.

        :param ids: IDs for the Events.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a list of dataclass for Events.
            Or return json data. Default is false.
        :return: Events information.
        """

        ids = enf_comma_separated(field="ids", value=ids)

        if fields is None:
            fields = const.EVENT_PUBLIC_FIELDS

        data = self.client.get_objects(
            ids=ids,
            fields=enf_comma_separated(field="fields", value=fields),
        )
        if return_json:
            return data
        else:
            return {
                event_id: Event.new_from_json_dict(item)
                for event_id, item in data.items()
            }



"""
    Model class for album.

    Refer: https://developers.facebook.com/docs/graph-api/reference/album/
"""

from dataclasses import dataclass
from typing import List, Optional

from dataclasses_json import config

from pyfacebook.models.base import BaseModel, field
from pyfacebook.models.extensions import Paging


@dataclass
class Album(BaseModel):
    """
    A class representing the Album.
    """

    id: Optional[str] = field(repr=True, compare=True)
    backdated_time: Optional[str] = field()
    backdated_time_granularity: Optional[str] = field()
    can_upload: Optional[bool] = field()
    count: Optional[int] = field()
    cover_photo: Optional[
        dict
    ] = (
        field()
    )  # TODO Refer: https://developers.facebook.com/docs/graph-api/reference/photo/
    created_time: Optional[str] = field()
    description: Optional[str] = field()
    event: Optional[
        dict
    ] = (
        field()
    )  # TODO Refer: https://developers.facebook.com/docs/graph-api/reference/event/
    _from: Optional[dict] = field(metadata=config(field_name="from"))
    link: Optional[str] = field()
    location: Optional[str] = field()
    name: Optional[str] = field(repr=True)
    place: Optional[
        dict
    ] = (
        field()
    )  # TODO Refer: https://developers.facebook.com/docs/graph-api/reference/place/
    privacy: Optional[str] = field()
    type: Optional[str] = field()
    updated_time: Optional[str] = field()


@dataclass
class AlbumResponse(BaseModel):
    """
    A class representing the result for albums edge.

    Refer: https://developers.facebook.com/docs/graph-api/reference/page/albums
    """

    data: List[Album] = field(repr=True, compare=True)
    paging: Optional[Paging] = field()



"""
    Model classes for instagram business.
"""

from dataclasses import dataclass
from typing import List, Optional, Union

from dataclasses_json import config

from pyfacebook.models.base import BaseModel, field
from pyfacebook.models.extensions import Paging


@dataclass
class IgBusUser(BaseModel):
    """
    A class representing the Business User.

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-user/
    """

    id: Optional[str] = field(repr=True, compare=True)
    ig_id: Optional[str] = field()
    biography: Optional[str] = field()
    followers_count: Optional[int] = field()
    follows_count: Optional[int] = field()
    media_count: Optional[int] = field()
    name: Optional[int] = field()
    profile_picture_url: Optional[str] = field()
    shopping_product_tag_eligibility: Optional[bool] = field()
    username: Optional[str] = field(repr=True)
    website: Optional[str] = field()


@dataclass
class IgBusMediaChild(BaseModel):
    """
    A class representing the child for media children.
    """

    id: Optional[str] = field(repr=True, compare=True)
    ig_id: Optional[str] = field()
    media_type: Optional[str] = field()
    media_url: Optional[str] = field(repr=True)
    owner: Optional[IgBusUser] = field()
    permalink: Optional[str] = field()
    shortcode: Optional[str] = field()
    thumbnail_url: Optional[str] = field()
    timestamp: Optional[str] = field()
    username: Optional[str] = field()


@dataclass
class IgBusMediaChildren(BaseModel):
    """
    A class representing the Children for the media.

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-media/children
    """

    data: Optional[List[IgBusMediaChild]] = field()


@dataclass
class IgBusMedia(BaseModel):
    """
    A class representing the Business Media.

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-media/
    """

    id: Optional[str] = field(repr=True, compare=True)
    ig_id: Optional[str] = field()
    caption: Optional[str] = field()
    children: Optional[IgBusMediaChildren] = field()
    comments_count: Optional[int] = field()
    is_comment_enabled: Optional[bool] = field()
    is_shared_to_feed: Optional[bool] = field()
    like_count: Optional[int] = field()
    media_product_type: Optional[str] = field()
    media_type: Optional[str] = field()
    media_url: Optional[str] = field(repr=True)
    owner: Optional[IgBusUser] = field()
    permalink: Optional[str] = field()
    shortcode: Optional[str] = field()
    thumbnail_url: Optional[str] = field()
    timestamp: Optional[str] = field()
    username: Optional[str] = field()
    # TODO Starting with v14.0, the video_title field will no longer be supported
    video_title: Optional[str] = field()


@dataclass
class IgBusMediaResponse(BaseModel):
    """
    A class representing the Business Medias response.

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-user/media
    """

    data: List[IgBusMedia] = field(repr=True, compare=True)
    paging: Optional[Paging] = field(repr=True)


@dataclass
class IgBusReply(BaseModel):
    """
    A class representing the Reply.
    Like a Comment.
    """

    id: Optional[str] = field(repr=True, compare=True)
    like_count: Optional[int] = field()
    media: Optional[IgBusMedia] = field()
    text: Optional[str] = field(repr=True)
    timestamp: Optional[str] = field()
    user: Optional[IgBusUser] = field()
    username: Optional[str] = field()
    _from: Optional[IgBusUser] = field(metadata=config(field_name="from"))


@dataclass
class IgBusReplies(BaseModel):
    """
    A class representing the Replies for comment.

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-comment/replies
    """

    data: Optional[List[IgBusReply]] = field()


@dataclass
class IgBusComment(IgBusReply):
    """
    A class representing the Business Comment.

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-comment/
    """

    replies: Optional[IgBusReplies] = field()


@dataclass
class IgBusCommentResponse(BaseModel):
    """
    A class representing the comments response.

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-media/comments
    """

    data: List[IgBusComment] = field(repr=True)
    paging: Optional[Paging] = field(repr=True)


@dataclass
class IgBusHashtag(BaseModel):
    """
    A class representing the hashtag.

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-hashtag/
    """

    id: Optional[str] = field(repr=True, compare=True)
    name: Optional[str] = field(repr=True)


@dataclass
class IgBusContainer(BaseModel):
    """
    A class representing the media container.

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-container
    """

    id: Optional[str] = field(repr=True, compare=True)
    status: Optional[str] = field()
    status_code: Optional[str] = field()


@dataclass
class IgBusPublishLimitConfig(BaseModel):
    """
    A class representing the content publish limit config.
    """

    quota_total: Optional[int] = field(repr=True)
    quota_duration: Optional[int] = field()


@dataclass
class IgBusPublishLimit(BaseModel):
    """
    A class representing the content publish limit.

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-user/content_publishing_limit
    """

    config: Optional[IgBusPublishLimitConfig] = field(repr=True)
    quota_usage: Optional[int] = field(repr=True)


@dataclass
class IgBusPublishLimitResponse(BaseModel):
    """
    A class representing the content publish limit response.

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-user/content_publishing_limit
    """

    data: List[IgBusPublishLimit] = field(repr=True)


@dataclass
class IgBusInsightValue(BaseModel):
    """
    A class representing the Instagram insight value info.

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-user/insights
    """

    value: Optional[Union[int, dict]] = field(repr=True)
    end_time: Optional[str] = field(repr=True)


@dataclass
class IgBusInsight(BaseModel):
    """
    A class representing the Instagram insight.

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-user/insights
    """

    id: Optional[str] = field()
    name: Optional[str] = field(repr=True)
    period: Optional[str] = field(repr=True)
    title: Optional[str] = field()
    description: Optional[str] = field()
    values: Optional[List[IgBusInsightValue]] = field()


@dataclass
class IgBusInsightsResponse(BaseModel):
    """
    A class representing the Instagram insights response.

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-user/insights
    """

    data: List[IgBusInsight] = field(repr=True)
    paging: Optional[Paging] = field()


@dataclass
class IgBusDiscoveryUserResponse(BaseModel):
    """
    A class representing the response for discovery user.

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-user/business_discovery#sample-response
    """

    business_discovery: Optional[IgBusUser] = field(repr=True)
    id: Optional[str] = field(repr=True)


@dataclass
class IgBusDiscoveryUserMedia(BaseModel):
    media: Optional[IgBusMediaResponse] = field(repr=True)
    id: Optional[str] = field()


@dataclass
class IgBusDiscoveryUserMediaResponse(BaseModel):
    """
    A class representing the response for discovery user.

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-user/business_discovery#sample-response
    """

    business_discovery: Optional[IgBusDiscoveryUserMedia] = field(repr=True)
    id: Optional[str] = field(repr=True)


@dataclass
class IgBusMentionedCommentResponse(BaseModel):
    """
    A class representing the response for mentioned comment.

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-user/mentioned_comment
    """

    mentioned_comment: Optional[IgBusReply] = field(repr=True)
    id: Optional[str] = field(repr=True)


@dataclass
class IgBusMentionedMediaResponse(BaseModel):
    """
    A class representing the response for mentioned media.

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-user/mentioned_media
    """

    mentioned_media: Optional[IgBusMedia] = field(repr=True)
    id: Optional[str] = field(repr=True)


@dataclass
class IgBusHashtagsResponse(BaseModel):
    """
    A class representing the hashtag.

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-user/recently_searched_hashtags
    """

    data: List[IgBusHashtag] = field(repr=True)
    paging: Optional[Paging] = field(repr=True)


@dataclass
class IgBusCatalog(BaseModel):
    """
    A class representing the catalog.

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-user/available_catalogs
    """

    catalog_id: Optional[str] = field(repr=True)
    catalog_name: Optional[str] = field(repr=True)
    shop_name: Optional[str] = field(repr=True)
    product_count: Optional[int] = field(repr=True)


@dataclass
class IgBusCatalogsResponse(BaseModel):
    """
    A class representing the catalog list response.

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-user/available_catalogs
    """

    data: List[IgBusCatalog] = field(repr=True)


@dataclass
class IgBusProductVariant(BaseModel):
    product_id: Optional[int] = field(repr=True)
    variant_name: Optional[str] = field(repr=True)


@dataclass
class IgBusProduct(BaseModel):
    """
    A class representing the catalog product.

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-user/catalog_product_search
    """

    product_id: Optional[int] = field(repr=True)
    merchant_id: Optional[int] = field(repr=True)
    product_name: Optional[str] = field(repr=True)
    image_url: Optional[str] = field()
    retailer_id: Optional[str] = field()
    review_status: Optional[str] = field()
    is_checkout_flow: Optional[bool] = field()
    product_variants: Optional[List[IgBusProductVariant]] = field()


@dataclass
class IgBusProductsResponse(BaseModel):
    """
    A class representing the catalog product list response.

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-user/catalog_product_search
    """

    data: List[IgBusProduct] = field(repr=True)
    paging: Optional[Paging] = field(repr=True)


@dataclass
class IgBusProductTag(BaseModel):
    """
    A class representing the product tag.

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-media/product_tags#ig-media-product-tags
    """

    product_id: Optional[int] = field(repr=True)
    merchant_id: Optional[int] = field(repr=True)
    name: Optional[str] = field(repr=True)
    price_string: Optional[str] = field()
    image_url: Optional[str] = field()
    review_status: Optional[str] = field()
    is_checkout: Optional[bool] = field()
    stripped_price_string: Optional[str] = field()
    string_sale_price_string: Optional[str] = field()
    x: Optional[float] = field()
    y: Optional[float] = field()


@dataclass
class IgBusProductTagsResponse(BaseModel):
    """
    A class representing the product tag list response.

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-media/product_tags#ig-media-product-tags
    """

    data: List[IgBusProductTag] = field(repr=True)


@dataclass
class IgBusProductAppeal(BaseModel):
    """
    A class representing the product appeal.

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-user/product_appeal
    """

    eligible_for_appeal: Optional[bool] = field(repr=True)
    product_id: Optional[int] = field(repr=True)
    review_status: Optional[str] = field(repr=True)


@dataclass
class IgBusProductAppealsResponse(BaseModel):
    """
    A class representing the product appeal list

    Refer: https://developers.facebook.com/docs/instagram-api/reference/ig-user/product_appeal#response-2
    """

    data: List[IgBusProductAppeal] = field(repr=True)



"""
    Base model
"""
from copy import deepcopy
from dataclasses import dataclass, field as base_field
from typing import (
    Dict,
    Type,
    TypeVar,
    Optional,
)

from dataclasses_json import (
    DataClassJsonMixin,
)
from dataclasses_json.core import Json

A = TypeVar("A", bound=DataClassJsonMixin)


@dataclass
class BaseModel(DataClassJsonMixin):
    @classmethod
    def new_from_json_dict(
        cls: Type[A], data: Optional[Dict], *, infer_missing=False
    ) -> Optional[A]:
        """
        Convert json dict to dataclass
        :param data: A json dict which will convert model class.
        :param infer_missing: if set True, will let missing field (not have default vale) to None
        :return: The data class
        """
        if not data:
            return None
        c = cls.from_dict(data, infer_missing=infer_missing)
        # save origin data
        cls._json = data
        return c

    def to_dict(self, encode_json=False, ignore_nan=True) -> Dict[str, Json]:
        """
        Convert dataclass to dict
        :param encode_json:
        :param ignore_nan: Is the result include None data.
        :return: dict
        """
        data = super().to_dict(encode_json=encode_json)
        if ignore_nan:
            data = dict_minus_none_values(data)
        return data


def dict_minus_none_values(obj: Optional[dict]):
    """
    Remove data dict where value is None
    :param obj: dict object
    :return: obj
    """
    if obj is None:
        return None

    keys = list(obj.keys())
    for key in keys:
        value = obj[key]
        if value is None:
            obj.pop(key)
        if isinstance(value, (list, tuple)):
            obj[key] = type(value)([dict_minus_none_values(v) for v in value])
        if isinstance(value, dict):
            obj[key] = dict_minus_none_values(value)
    return deepcopy(obj)


def field(default=None, repr=False, compare=False, **kwargs):
    kwargs.update(
        {
            "default": default,
            "repr": repr,
            "compare": compare,
        }
    )
    return base_field(**kwargs)



"""
    Model class for Conversation.

    Refer: https://developers.facebook.com/docs/graph-api/reference/conversation
"""
from dataclasses import dataclass
from typing import List, Optional

from pyfacebook.models.base import BaseModel, field


@dataclass
class CvUser(BaseModel):
    """
    A class representing the User in conversation.
    """

    id: Optional[str] = field(repr=True, compare=True)
    name: Optional[str] = field(repr=True)
    email: Optional[str] = field()


@dataclass
class Participants(BaseModel):
    """
    A class representing the Participant.
    """

    data: Optional[List[CvUser]] = field(repr=True)


@dataclass
class Senders(BaseModel):
    """
    A class representing the Sender.
    """

    data: Optional[List[CvUser]] = field(repr=True)


@dataclass
class Conversation(BaseModel):
    """
    A class representing the Conversation.
    """

    id: Optional[str] = field(repr=True, compare=True)
    link: Optional[str] = field()
    snippet: Optional[str] = field()
    scoped_thread_key: Optional[str] = field()
    updated_time: Optional[str] = field(repr=True)
    message_count: Optional[int] = field()
    unread_count: Optional[int] = field()
    participants: Optional[Participants] = field()
    senders: Optional[Senders] = field()
    can_reply: Optional[bool] = field()
    is_subscribed: Optional[bool] = field()



"""
    tests for base graph api.
"""
import pytest
import requests
import responses

from pyfacebook import GraphAPI, LibraryError, FacebookError


def test_api_initial():
    # test invalid version
    with pytest.raises(LibraryError):
        GraphAPI(version="ace")
    # test not support version
    with pytest.raises(LibraryError):
        GraphAPI(version="1.0")

    # test without access token
    with pytest.raises(LibraryError):
        GraphAPI()

    # use oauth flow
    api = GraphAPI(
        app_id="app id",
        app_secret="app secret",
        oauth_flow=True,
        base_url="https://graph.facebook.com/",
    )


def test_sleep_seconds_mapping(pubg_api):
    r = pubg_api._build_sleep_seconds_resource({1: 1, 2: 2})
    assert r[0].seconds == 1

    r = pubg_api._build_sleep_seconds_resource(None)
    assert r is None


def test_request_error(pubg_api):
    with pytest.raises(LibraryError):
        with responses.RequestsMock() as m:
            m.add(
                "GET", "https://graph.facebook.com/", body=requests.HTTPError("Wrong")
            )

            pubg_api._request(
                url="https://graph.facebook.com/",
            )


def test_parse_response(pubg_api):
    resp = requests.Response()
    resp.headers["Content-Type"] = "multipart/form-data"

    with pytest.raises(LibraryError):
        pubg_api._parse_response(response=resp)


def test_graph_error(helpers, pubg_api):
    data = helpers.load_json("testdata/base/error_data.json")

    with pytest.raises(FacebookError):
        pubg_api._check_graph_error(data)


def test_get_object(helpers, pubg_api):
    obj_id = "20531316728"

    # test default field
    with responses.RequestsMock() as m:
        m.add(
            method="GET",
            url=f"https://graph.facebook.com/{pubg_api.version}/{obj_id}",
            json=helpers.load_json("testdata/base/object_default.json"),
        )

        res = pubg_api.get_object(object_id=obj_id, fields="")
        assert res["id"] == obj_id

    # test more fields
    with responses.RequestsMock() as m:
        m.add(
            method="GET",
            url=f"https://graph.facebook.com/{pubg_api.version}/{obj_id}",
            json=helpers.load_json("testdata/base/object_data.json"),
        )

        res = pubg_api.get_object(
            object_id=obj_id,
            fields="id,about,can_checkin,category,category_list,checkins,contact_address,cover,current_location,description,description_html,display_subtext,emails,engagement,fan_count,founded,general_info,global_brand_page_name,global_brand_root_id,link,name,phone,picture,rating_count,single_line_address,start_info,talking_about_count,username,verification_status,website,were_here_count,whatsapp_number",
            appsecret_proof="xxxx",
        )
        assert res["id"] == obj_id
        assert res["engagement"]["count"] == 209924887


def test_get_objects(helpers, pubg_api):
    ids = "20531316728,19292868552"

    with responses.RequestsMock() as m:
        m.add(
            method="GET",
            url=f"https://graph.facebook.com/{pubg_api.version}",
            json=helpers.load_json("testdata/base/objects_default.json"),
        )

        res = pubg_api.get_objects(ids=ids)
        assert len(res) == 2
        assert res["20531316728"]["id"] == "20531316728"

        res = pubg_api.get_objects(
            ids=ids,
            fields="id,name",
            appsecret_proof="xxxx",
        )
        assert len(res) == 2


def test_get_connection(helpers, pubg_api):
    api = GraphAPI(access_token="token")
    obj_id = "19292868552"

    # image
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{pubg_api.version}/{obj_id}/picture",
            body=helpers.load_file_binary("testdata/base/19292868552-picture.png"),
            content_type="image/png",
        )
        res = api.get_connection(
            object_id=obj_id,
            connection="picture",
            redirect=0,
        )
        assert res["content-type"] == "image/png"

    # data
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{pubg_api.version}/{obj_id}/picture",
            json=helpers.load_json("testdata/base/connecion_data.json"),
        )

        res = api.get_connection(object_id=obj_id, connection="picture")
        assert res["data"]["height"] == 50


def test_get_full_connections(helpers):
    obj_id = "19292868552"

    api = GraphAPI(access_token="token", version="v11.0")
    # test with count
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url="https://graph.facebook.com/v11.0/19292868552/feed",
            json=helpers.load_json("testdata/base/full_connecions_p1.json"),
        )

        feed = api.get_full_connections(
            object_id=obj_id,
            connection="feed",
            count=3,
            limit=5,
            fields="created_time,id,message",
        )
        assert len(feed["data"]) == 3

    # test with no next
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url="https://graph.facebook.com/v11.0/19292868552/feed",
            json=helpers.load_json("testdata/base/full_connecions_p1.json"),
        )
        m.add(
            method=responses.GET,
            url="https://graph.facebook.com/v11.0/19292868552/feed",
            json=helpers.load_json("testdata/base/full_connecions_p2.json"),
        )

        feed = api.get_full_connections(
            object_id=obj_id,
            connection="feed",
            count=None,
            limit=5,
            fields="created_time,id,message",
        )
        assert len(feed["data"]) == 8


def test_discovery_user_media(helpers):
    username = "facebook"

    api = GraphAPI(access_token="token", instagram_business_id="id")
    # test with count
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/id",
            json=helpers.load_json("testdata/base/discovery_medias_p1.json"),
        )

        media = api.discovery_user_media(
            username=username,
            fields="comments_count,id,like_count,media_type,media_url,permalink,timestamp",
            count=3,
            limit=5,
        )
        assert len(media["data"]) == 3

    # test with no next
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/id",
            json=helpers.load_json("testdata/base/discovery_medias_p1.json"),
        )

        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/id",
            json=helpers.load_json("testdata/base/discovery_medias_p2.json"),
        )

        media = api.discovery_user_media(
            username=username,
            fields="comments_count,id,like_count,media_type,media_url,permalink,timestamp",
            count=None,
            limit=5,
        )
        assert len(media["data"]) == 10


def test_post_object(helpers, user_api):
    obj_id = "123456789"
    with responses.RequestsMock() as m:
        m.add(
            method="POST",
            url=f"https://graph.facebook.com/{user_api.version}/{obj_id}/comments",
            json={"success": True},
        )

        res = user_api.post_object(
            object_id=obj_id,
            connection="comments",
            data={"message": "data"},
        )
        assert res["success"]


def test_delete_object(helpers, user_api):
    obj_id = "123456789"
    with responses.RequestsMock() as m:
        m.add(
            method="DELETE",
            url=f"https://graph.facebook.com/{user_api.version}/{obj_id}",
            json={"success": True},
        )

        res = user_api.delete_object(
            object_id=obj_id,
        )
        assert res["success"]


def test_oauth_flow(helpers):
    with pytest.raises(LibraryError):
        api = GraphAPI(access_token="token")
        api.get_authorization_url()

    api = GraphAPI(app_id="id", app_secret="secret", oauth_flow=True)

    # test get authorization url
    _, state = api.get_authorization_url()

    # if user give authorize.
    resp = "https://localhost/?code=code&state=PyFacebook#_=_"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.POST,
            url=api.EXCHANGE_ACCESS_TOKEN_URL,
            json=helpers.load_json("testdata/base/long_term_token.json"),
        )

        r = api.exchange_user_access_token(response=resp)
        assert r["access_token"] == "token"


def test_exchange_token(helpers):
    api = GraphAPI(access_token="token")

    page_id = "19292868552"

    # test page access token
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{page_id}",
            json={"id": "19292868552", "access_token": "token"},
        )
        token = api.exchange_page_access_token(page_id=page_id)
        assert token == "token"
    # test can not exchange page access token
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{page_id}",
            json={"id": "19292868552"},
        )
        with pytest.raises(LibraryError):
            api.exchange_page_access_token(page_id=page_id)

    # test exchange long-lived user token
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/oauth/access_token",
            json={
                "access_token": "token",
                "token_type": "bearer",
                "expires_in": 5184000,
            },
        )

        res = api.exchange_long_lived_user_access_token()
        assert res["access_token"] == "token"

    user_id = "123456"
    # test exchange long-lived page token
    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{user_id}/accounts",
            json=helpers.load_json("testdata/base/user_accounts_token.json"),
        )

        res = api.exchange_long_lived_page_access_token(user_id=user_id)
        assert len(res["data"]) == 1
        assert res["data"][0]["access_token"] == "access_token"


def test_get_app_token():
    api = GraphAPI(app_id="id", app_secret="secret", access_token="token")

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/oauth/access_token",
            json={"access_token": "access_token", "token_type": "bearer"},
        )

        data = api.get_app_token()
        assert data["access_token"] == "access_token"

        # initial with application_only_auth
        GraphAPI(app_id="id", app_secret="secret", application_only_auth=True)


def test_debug_token(helpers, pubg_api):
    input_token = "token"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{pubg_api.version}/debug_token",
            json=helpers.load_json("testdata/base/token_info.json"),
        )

        res = pubg_api.debug_token(input_token=input_token)
        assert res["data"]["type"] == "USER"
        assert res["data"]["is_valid"]



"""
    Tests for live video.
"""

import responses


def test_get_info(helpers, fb_api):
    lv_id = "10158276101223553"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{lv_id}",
            json=helpers.load_json(
                "testdata/facebook/apidata/live_videos/live_video_fields.json"
            ),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{lv_id}",
            json=helpers.load_json(
                "testdata/facebook/apidata/live_videos/live_video_default.json"
            ),
        )

        live_video = fb_api.live_video.get_info(live_video_id=lv_id)
        assert live_video.id == lv_id
        assert live_video.status == "VOD"

        live_video_json = fb_api.live_video.get_info(
            live_video_id=lv_id,
            fields="id,title,status,embed_html",
            return_json=True,
        )
        assert live_video_json["id"] == lv_id
        assert live_video_json["title"] == "F8 Refresh 2021 Sessions"


def test_get_batch(helpers, fb_api):
    lv_ids = ["10158276101223553", "10158275863243553"]

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}",
            json=helpers.load_json(
                "testdata/facebook/apidata/live_videos/live_videos_fields.json"
            ),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}",
            json=helpers.load_json(
                "testdata/facebook/apidata/live_videos/live_videos_default.json"
            ),
        )

        live_videos = fb_api.live_video.get_batch(ids=lv_ids)
        assert live_videos[lv_ids[0]].id == lv_ids[0]

        live_videos_json = fb_api.live_video.get_batch(
            ids=lv_ids,
            fields="id,title,status,embed_html",
            return_json=True,
        )
        assert live_videos_json[lv_ids[1]]["id"] == lv_ids[1]



"""
    Apis for media.
"""
from typing import Dict, Optional, Union

import pyfacebook.utils.constant as const
from pyfacebook.api.base_resource import BaseResource
from pyfacebook.models.ig_business_models import (
    IgBusMedia,
    IgBusCommentResponse,
    IgBusMediaChildren,
    IgBusInsightsResponse,
    IgBusProductTagsResponse,
)
from pyfacebook.utils.params_utils import enf_comma_separated


class IGBusinessMedia(BaseResource):
    def get_info(
        self,
        media_id: str,
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[IgBusMedia, dict]:
        """
        Get information about a Business media.

        :param media_id: ID for Media.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for IgBusMedia.
            Or return json data. Default is false.
        :return: Business media information.
        """
        if fields is None:
            fields = const.IG_BUSINESS_MEDIA_PUBLIC_FIELDS

        data = self.client.get_object(
            object_id=media_id,
            fields=enf_comma_separated(field="fields", value=fields),
        )
        if return_json:
            return data
        else:
            return IgBusMedia.new_from_json_dict(data=data)

    def get_batch(
        self,
        ids: Optional[Union[str, list, tuple]],
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[Dict[str, IgBusMedia], dict]:
        """
        Get batch business media information by ids

        :param ids: IDs for the medias.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for IgBusMedia.
            Or return json data. Default is false.
        :return: Business medias information.
        """
        ids = enf_comma_separated(field="ids", value=ids)

        if fields is None:
            fields = const.IG_BUSINESS_MEDIA_PUBLIC_FIELDS

        data = self.client.get_objects(
            ids=ids, fields=enf_comma_separated(field="fields", value=fields)
        )
        if return_json:
            return data
        else:
            return {
                media_id: IgBusMedia.new_from_json_dict(item)
                for media_id, item in data.items()
            }

    def get_comments(
        self,
        media_id: str,
        fields: Optional[Union[str, list, tuple]] = None,
        count: Optional[int] = 10,
        limit: Optional[int] = 10,
        return_json: bool = False,
    ) -> Union[IgBusCommentResponse, dict]:
        """
        Get list of IG Comments on an IG Media object.

        :param media_id: ID for the media.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param count: The total count for you to get data.
        :param limit: Each request retrieve objects count.
            It should no more than 50. Default is None will use api default limit.
        :param return_json: Set to false will return a dataclass for IgBusCommentResponse.
            Or return json data. Default is false.
        :return: Comments response information.
        """

        if fields is None:
            fields = const.IG_BUSINESS_COMMENT_PUBLIC_FIELDS

        data = self.client.get_full_connections(
            object_id=media_id,
            connection="comments",
            count=count,
            limit=limit,
            fields=enf_comma_separated(field="fields", value=fields),
        )
        if return_json:
            return data
        else:
            return IgBusCommentResponse.new_from_json_dict(data)

    def get_children(
        self,
        media_id: str,
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[IgBusMediaChildren, dict]:
        """
        Get list of IG Media objects on an album IG Media object.

        :param media_id: ID for the album media.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for IgBusMediaChildren.
            Or return json data. Default is false.
        :return: Media children response information.
        """

        if fields is None:
            fields = const.IG_BUSINESS_MEDIA_CHILDREN_PUBLIC_FIELDS

        data = self.client.get_connection(
            object_id=media_id,
            connection="children",
            fields=enf_comma_separated(field="fields", value=fields),
        )
        if return_json:
            return data
        else:
            return IgBusMediaChildren.new_from_json_dict(data)

    def get_insights(
        self,
        media_id: str,
        metric: Union[str, list, tuple],
        return_json: bool = False,
    ) -> Union[IgBusInsightsResponse, dict]:
        """
        Get insights data on a media

        :param media_id: ID for the media.
        :param metric: A comma-separated list of Metrics you want returned.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for IgBusInsightsResponse.
            Or return json data. Default is false.
        :return: Media insights response information.
        """
        data = self.client.get_connection(
            object_id=media_id,
            connection="insights",
            metric=enf_comma_separated(field="metric", value=metric),
        )

        if return_json:
            return data
        else:
            return IgBusInsightsResponse.new_from_json_dict(data)

    def get_product_tags(
        self, media_id: str, return_json: bool = False
    ) -> Union[IgBusProductTagsResponse, dict]:
        """
        Get a collection of product tags on an IG Media.
        :param media_id: ID for the media.
        :param return_json: Set to false will return a dataclass for IgBusProductTagsResponse.
            Or return json data. Default is false.
        :return: Media product tags
        """

        data = self.client.get_connection(
            object_id=media_id,
            connection="product_tags",
        )
        if return_json:
            return data
        else:
            return IgBusProductTagsResponse.new_from_json_dict(data)



"""
    Apis for comment.
"""
from typing import Dict, Optional, Union

import pyfacebook.utils.constant as const
from pyfacebook.api.base_resource import BaseResource
from pyfacebook.models.ig_business_models import IgBusComment, IgBusReply, IgBusReplies
from pyfacebook.utils.params_utils import enf_comma_separated


class IGBusinessComment(BaseResource):
    def get_info(
        self,
        comment_id: str,
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[IgBusComment, dict]:
        """
        Get information about a Business comment.

        :param comment_id: ID for Comment.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for IgBusComment.
            Or return json data. Default is false.
        :return: Business comment information.
        """
        if fields is None:
            fields = const.IG_BUSINESS_MEDIA_PUBLIC_FIELDS

        data = self.client.get_object(
            object_id=comment_id,
            fields=enf_comma_separated(field="fields", value=fields),
        )
        if return_json:
            return data
        else:
            return IgBusComment.new_from_json_dict(data=data)

    def get_batch(
        self,
        ids: Optional[Union[str, list, tuple]],
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[Dict[str, IgBusComment], dict]:
        """
        Get batch business comment information by ids

        :param ids: IDs for the comments.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dict of dataclass for IgBusComment.
            Or return json data. Default is false.
        :return: Business medias information.
        """
        ids = enf_comma_separated(field="ids", value=ids)

        if fields is None:
            fields = const.IG_BUSINESS_COMMENT_PUBLIC_FIELDS

        data = self.client.get_objects(
            ids=ids, fields=enf_comma_separated(field="fields", value=fields)
        )
        if return_json:
            return data
        else:
            return {
                comment_id: IgBusComment.new_from_json_dict(item)
                for comment_id, item in data.items()
            }

    def get_replies(
        self,
        comment_id: str,
        fields: Optional[Union[str, list, tuple]] = None,
        count: Optional[int] = 10,
        limit: Optional[int] = 10,
        return_json: bool = False,
    ) -> Union[IgBusReplies, dict]:
        """
        Getting All Replies (Comments) on a Comment

        :param comment_id: ID for the comment.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param count: The total count for you to get data.
        :param limit: Each request retrieve objects count.
            It should no more than 100. Default is None will use api default limit.
        :param return_json: Set to false will return a dataclass for IgBusReplies.
            Or return json data. Default is false.
        :return: Comment replies response information.
        """

        if fields is None:
            fields = const.IG_BUSINESS_REPLY_PUBLIC_FIELDS

        data = self.client.get_full_connections(
            object_id=comment_id,
            connection="replies",
            fields=enf_comma_separated(field="fields", value=fields),
            count=count,
            limit=limit,
        )
        if return_json:
            return data
        else:
            return IgBusReplies.new_from_json_dict(data)


class IGBusinessReply(BaseResource):
    def get_info(
        self,
        reply_id: str,
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[IgBusReply, dict]:
        """
        Get information about a Business reply.

        :param reply_id: ID for reply.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for IgBusReply.
            Or return json data. Default is false.
        :return: Business reply information.
        """
        if fields is None:
            fields = const.IG_BUSINESS_REPLY_PUBLIC_FIELDS

        data = self.client.get_object(
            object_id=reply_id,
            fields=enf_comma_separated(field="fields", value=fields),
        )
        if return_json:
            return data
        else:
            return IgBusComment.new_from_json_dict(data=data)

    def get_batch(
        self,
        ids: Optional[Union[str, list, tuple]],
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[Dict[str, IgBusReply], dict]:
        """
        Get batch business replies information by ids

        :param ids: IDs for the replies.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dict of dataclass for IgBusReply.
            Or return json data. Default is false.
        :return: Business replies information.
        """
        ids = enf_comma_separated(field="ids", value=ids)

        if fields is None:
            fields = const.IG_BUSINESS_REPLY_PUBLIC_FIELDS

        data = self.client.get_objects(
            ids=ids, fields=enf_comma_separated(field="fields", value=fields)
        )
        if return_json:
            return data
        else:
            return {
                reply_id: IgBusReply.new_from_json_dict(item)
                for reply_id, item in data.items()
            }



from .base import BaseModel
from .business import Business
from .user import User, UserExperience, UserAgeRange
from .page import Page, PagesResponse, SearchPage, SearchPagesResponse
from .post import Post, FeedResponse
from .group import Group
from .event import Event
from .photo import Photo, PhotosResponse
from .album import Album, AlbumResponse
from .video import Video, VideosResponse
from .live_video import LiveVideo, LiveVideosResponse
from .comment import Comment, CommentsResponse
from .conversation import Conversation
from .message import Message
from .ig_business_models import (
    IgBusUser,
    IgBusMedia,
    IgBusMediaResponse,
    IgBusMediaChildren,
    IgBusReply,
    IgBusComment,
    IgBusHashtag,
    IgBusContainer,
    IgBusPublishLimit,
    IgBusPublishLimitResponse,
    IgBusInsight,
    IgBusInsightsResponse,
    IgBusDiscoveryUserResponse,
    IgBusDiscoveryUserMediaResponse,
    IgBusMentionedCommentResponse,
    IgBusMentionedMediaResponse,
    IgBusHashtagsResponse,
    IgBusCommentResponse,
)
from .ig_basic_models import (
    IgBasicUser,
    IgBasicMediaChildren,
    IgBasicMedia,
    IgBasicMediaResponse,
)



"""
    Tests for post api
"""

import responses


def test_post_info(helpers, fb_api):
    post_id = "175154750010052_424924701699721"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{post_id}",
            json=helpers.load_json(
                "testdata/facebook/apidata/posts/single_post_data.json"
            ),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{post_id}",
            json=helpers.load_json(
                "testdata/facebook/apidata/posts/single_post_data_fields.json"
            ),
        )

        post = fb_api.post.get_info(post_id=post_id)
        assert post.id == post_id

        post_json = fb_api.post.get_info(
            post_id=post_id, fields="id,created_time,message", return_json=True
        )
        assert post_json["id"] == post_id


def test_batch_posts(helpers, fb_api):
    ids = ["19292868552_10158347294088553", "19292868552_10158345715593553"]

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}",
            json=helpers.load_json(
                "testdata/facebook/apidata/posts/multi_posts_data.json"
            ),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}",
            json=helpers.load_json(
                "testdata/facebook/apidata/posts/multi_posts_data_fields.json"
            ),
        )

        data = fb_api.post.get_batch(ids=ids)
        assert ids[0] in data.keys()

        data_json = fb_api.post.get_batch(
            ids=ids, fields="id,message,created_time,updated_time", return_json=True
        )
        assert data_json[ids[0]]["id"] == ids[0]


def test_get_comments(helpers, fb_api):
    post_id = "19292868552_10158407654328553"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{post_id}/comments",
            json=helpers.load_json(
                "testdata/facebook/apidata/comments/comments_p1.json"
            ),
        )
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{fb_api.version}/{post_id}/comments",
            json=helpers.load_json(
                "testdata/facebook/apidata/comments/comments_p2.json"
            ),
        )

        comments = fb_api.post.get_comments(
            object_id=post_id,
            count=None,
            limit=10,
        )
        assert len(comments.data) == 15
        assert comments.summary.total_count == 18

        comments_json = fb_api.post.get_comments(
            object_id=post_id, count=5, limit=5, return_json=True
        )
        assert len(comments_json["data"]) == 5



"""
    Client for Instagram Basic Display API.
"""

from pyfacebook.api.base_client import BaseBasicDisplayApi
from pyfacebook.api.instagram_basic import resource as rs


class IGBasicDisplayApi(BaseBasicDisplayApi):
    """
    Api class for Instagram basic display api
    """

    user = rs.IGBasicUser()
    media = rs.IGBasicMedia()



"""
    Models for application.

    Refer: https://developers.facebook.com/docs/graph-api/reference/application
"""

from dataclasses import dataclass
from typing import Any, List, Optional

from pyfacebook.models.base import BaseModel, field
from pyfacebook.models.extensions import Paging


@dataclass
class ApplicationObjectStoreURLs(BaseModel):
    """
    A class representing the Application Object Store URLs.

    Refer: https://developers.facebook.com/docs/graph-api/reference/application-object-store-urls/
    """

    amazon_app_store: Optional[str] = field()
    fb_canvas: Optional[str] = field()
    fb_gameroom: Optional[str] = field()
    google_play: Optional[str] = field(repr=True)
    instant_game: Optional[str] = field()
    itunes: Optional[str] = field(repr=True)
    itunes_ipad: Optional[str] = field()
    windows_10_store: Optional[str] = field()


@dataclass
class ApplicationRestrictionInfo(BaseModel):
    """
    A class representing the Application Restriction Info.

    Refer: https://developers.facebook.com/docs/graph-api/reference/application-restriction-info/
    """

    age: Optional[str] = field(repr=True)
    age_distribution: Optional[str] = field()
    location: Optional[str] = field()
    type: Optional[str] = field(repr=True)


@dataclass
class Application(BaseModel):
    """
    A class representing the Application.
    """

    id: Optional[str] = field(repr=True, compare=True)
    aam_rules: Optional[str] = field()
    an_ad_space_limit: Optional[int] = field()
    an_platforms: Optional[List[str]] = field()
    app_domains: Optional[List[str]] = field()
    app_events_config: Optional[dict] = field()  # TODO
    app_install_tracked: Optional[bool] = field()
    app_name: Optional[str] = field()
    app_signals_binding_ios: Optional[List[dict]] = field()  # TODO
    app_type: Optional[int] = field()
    auth_dialog_data_help_url: Optional[str] = field()
    auth_dialog_headline: Optional[str] = field()
    auth_dialog_perms_explanation: Optional[str] = field()
    auth_referral_default_activity_privacy: Optional[str] = field()
    auth_referral_enabled: Optional[int] = field()
    auth_referral_extended_perms: Optional[List[str]] = field()
    auth_referral_friend_perms: Optional[List[str]] = field()
    auth_referral_response_type: Optional[str] = field()
    auth_referral_user_perms: Optional[List[str]] = field()
    business: Optional[
        dict
    ] = (
        field()
    )  # TODO Refer: https://developers.facebook.com/docs/marketing-api/reference/business/
    canvas_fluid_height: Optional[bool] = field()
    canvas_fluid_width: Optional[int] = field()
    canvas_url: Optional[str] = field()
    category: Optional[str] = field()
    client_config: Optional[dict] = field()
    company: Optional[str] = field()
    configured_ios_sso: Optional[str] = field()
    contact_email: Optional[str] = field()
    created_time: Optional[str] = field()
    creator_uid: Optional[str] = field()
    daily_active_users: Optional[int] = field()
    daily_active_users_rank: Optional[int] = field()
    deauth_callback_url: Optional[str] = field()
    default_share_mode: Optional[str] = field()
    description: Optional[str] = field()
    financial_id: Optional[str] = field()
    hosting_url: Optional[str] = field()
    icon_url: Optional[str] = field()
    ios_bundle_id: Optional[List[str]] = field()
    ios_supports_native_proxy_auth_flow: Optional[bool] = field()
    ios_supports_system_auth: Optional[bool] = field()
    ipad_app_store_id: Optional[str] = field()
    iphone_app_store_id: Optional[str] = field()
    is_eligible_for_value_optimization: Optional[bool] = field()
    latest_sdk_version: Optional[Any] = field()  # TODO
    link: Optional[str] = field(repr=True)
    logging_token: Optional[str] = field()
    logo_url: Optional[str] = field()
    migrations: Optional[dict] = field()
    mobile_profile_section_url: Optional[str] = field()
    mobile_web_url: Optional[str] = field()
    monthly_active_users: Optional[int] = field()
    monthly_active_users_rank: Optional[int] = field()
    name: Optional[str] = field()
    namespace: Optional[str] = field()
    object_store_urls: Optional[ApplicationObjectStoreURLs] = field()
    page_tab_default_name: Optional[str] = field()
    page_tab_url: Optional[str] = field()
    photo_url: Optional[str] = field()
    privacy_policy_url: Optional[str] = field()
    profile_section_url: Optional[str] = field()
    property_id: Optional[str] = field()
    real_time_mode_devices: Optional[List[str]] = field()
    restrictions: Optional[ApplicationRestrictionInfo] = field()
    restrictive_data_filter_params: Optional[str] = field()
    secure_canvas_url: Optional[str] = field()
    secure_page_tab_url: Optional[str] = field()
    server_ip_whitelist: Optional[str] = field()
    social_discovery: Optional[int] = field()
    subcategory: Optional[str] = field()
    suggested_events_setting: Optional[str] = field()
    supported_platforms: Optional[List[str]] = field()
    terms_of_service_url: Optional[str] = field()
    url_scheme_suffix: Optional[str] = field()
    user_support_email: Optional[str] = field()
    user_support_url: Optional[str] = field()
    website_url: Optional[str] = field()
    weekly_active_users: Optional[int] = field()


@dataclass
class ApplicationTestAccount(BaseModel):
    """
    A class representing the test account for the application.

    Refer: https://developers.facebook.com/docs/graph-api/reference/test-account/
    """

    id: Optional[str] = field(repr=True, compare=True)
    login_url: Optional[str] = field(repr=True)
    access_token: Optional[str] = field()


@dataclass
class ApplicationAccountsResponse(BaseModel):
    """
    A class representing the application edge accounts response.

    Refer: https://developers.facebook.com/docs/graph-api/reference/application/accounts/
    """

    data: Optional[List[ApplicationTestAccount]] = field(repr=True)
    paging: Optional[Paging] = field()



"""
    Expose for outside
"""

from pyfacebook.api.facebook.resource.application import FacebookApplication
from pyfacebook.api.facebook.resource.business import FacebookBusiness
from pyfacebook.api.facebook.resource.user import FacebookUser
from pyfacebook.api.facebook.resource.page import FacebookPage
from pyfacebook.api.facebook.resource.group import FacebookGroup
from pyfacebook.api.facebook.resource.event import FacebookEvent
from pyfacebook.api.facebook.resource.post import FacebookPost
from pyfacebook.api.facebook.resource.photo import FacebookPhoto
from pyfacebook.api.facebook.resource.album import FacebookAlbum
from pyfacebook.api.facebook.resource.video import FacebookVideo
from pyfacebook.api.facebook.resource.live_video import FacebookLiveVideo
from pyfacebook.api.facebook.resource.comment import FacebookComment
from pyfacebook.api.facebook.resource.conversation import FacebookConversation
from pyfacebook.api.facebook.resource.message import FacebookMessage



from pyfacebook.api.instagram_business.resource.user import IGBusinessUser
from pyfacebook.api.instagram_business.resource.media import IGBusinessMedia
from pyfacebook.api.instagram_business.resource.comment import (
    IGBusinessComment,
    IGBusinessReply,
)
from pyfacebook.api.instagram_business.resource.hashtag import IGBusinessHashtag
from pyfacebook.api.instagram_business.resource.container import IGBusinessContainer



from requests.models import CaseInsensitiveDict
from pyfacebook import RateLimit, PercentSecond


def test_parse_headers():
    headers = CaseInsensitiveDict({"key": "None"})
    assert RateLimit.parse_headers(headers, key="key") is None
    assert RateLimit.parse_headers(headers, key="key-null") is None


def test_app_limit():
    headers = CaseInsensitiveDict(
        {"x-app-usage": '{"call_count":91,"total_cputime":15,"total_time":12}'}
    )
    r = RateLimit()
    r.set_limit(headers)

    assert r.get_limit().total_cputime == 15
    assert r.get_limit().max_percent() == 91
    assert r.get_max_percent() == 91
    assert r.get_sleep_seconds() == 0

    mapping = [PercentSecond(10, 1), PercentSecond(20, 2)]
    assert r.get_sleep_seconds(sleep_data=mapping) == 600

    headers = CaseInsensitiveDict(
        {"x-app-usage": '{"call_count":16,"total_cputime":15,"total_time":12}'}
    )
    r.set_limit(headers)
    assert r.get_sleep_seconds(sleep_data=mapping) == 2


def test_business_limit():
    headers = CaseInsensitiveDict(
        {
            "x-business-use-case-usage": '{"112130216863063":[{"type":"pages","call_count":1,"total_cputime":1,"total_time":1,"estimated_time_to_regain_access":0}]}'
        }
    )

    r = RateLimit()
    r.set_limit(headers)

    assert r.get_limit("112130216863063", "pages").call_count == 1



"""
    This is default fields for to retrieve data from Facebook graph API.
"""

# =============================================
# Follows is Facebook graph fields to get data
# =============================================

BUSINESS_PUBLIC_FIELDS = [
    "id",
    "name",
    "link",
    "created_time",
    "updated_time",
    "verification_status",
    "profile_picture_uri",
]

USER_PUBLIC_FIELDS = [
    "id",
    "first_name",
    "last_name",
    "middle_name",
    "name",
    "name_format",
    "picture",
    "short_name",
]

PAGE_PUBLIC_FIELDS = [
    "id",
    "about",
    "category",
    "category_list",
    "cover",
    "description",
    "description_html",
    "display_subtext",
    "emails",
    "engagement",
    "fan_count",
    "global_brand_page_name",
    "global_brand_root_id",
    "link",
    "name",
    "phone",
    "picture",
    "username",
    "verification_status",
    "website",
]

SEARCH_PAGES_PUBLIC_FIELDS = ["id", "name", "location", "link", "verification_status"]

GROUP_PUBLIC_FIELDS = [
    "id",
    "name",
    "description",
    "created_time",
    "member_count",
]

EVENT_PUBLIC_FIELDS = [
    "id",
    "name",
    "description",
    "start_time",
    "end_time",
    "place",
    "event_times",
]

ATTACHMENT_FIELDS = [
    "description",
    "description_tags",
    "media",
    "media_type",
    "target",
    "title",
    "type",
    "unshimmed_url",
    "url",
    "subattachments",
]

POST_PUBLIC_FIELDS = [
    "id",
    f"attachments{{{','.join(ATTACHMENT_FIELDS)}}}",
    "created_time",
    "full_picture",
    "icon",
    "message",
    "message_tags",
    "permalink_url",
    "picture",
    "shares",
    "status_type",
    "updated_time",
]

POST_CONNECTIONS_SUMMERY_FIELDS = [
    "comments.filter(stream).summary(true).limit(0)",
    "reactions.summary(true).limit(0)",
    "reactions.type(LIKE).limit(0).summary(total_count).as(like)",
    "reactions.type(LOVE).limit(0).summary(total_count).as(love)",
    "reactions.type(WOW).limit(0).summary(total_count).as(wow)",
    "reactions.type(HAHA).limit(0).summary(total_count).as(haha)",
    "reactions.type(SAD).limit(0).summary(total_count).as(sad)",
    "reactions.type(ANGRY).limit(0).summary(total_count).as(angry)",
]

PHOTO_PUBLIC_FIELDS = ["id", "created_time", "name", "updated_time", "link"]

ALBUM_PUBLIC_FIELDS = ["id", "created_time", "count", "name", "updated_time", "link"]

APPLICATION_PUBLIC_FIELDS = [
    "id",
    "category",
    "description",
    "link",
    "name",
    "namespace",
]

APPLICATION_ACCOUNT_PUBLIC_FIELDS = ["id", "login_url"]

VIDEO_PUBLIC_FIELDS = [
    "id",
    "created_time",
    "description",
    "title",
    "published",
    "length",
    "permalink_url",
    "updated_time",
]

LIVE_VIDEO_PUBLIC_FIELDS = [
    "id",
    "creation_time",
    "description",
    "embed_html",
    "live_views",
    "permalink_url",
    "secure_stream_url",
    "status",
    "stream_url",
    "title",
]

COMMENT_PUBLIC_FIELDS = [
    "id",
    "comment_count",
    "created_time",
    "like_count",
    "message",
    "permalink_url",
]

CONVERSATION_FIELDS = [
    "id",
    "link",
    "message_count",
    "snippet",
    "unread_count",
    "updated_time",
]

MESSAGE_FIELDS = [
    "id",
    "created_time",
    "from",
    "to",
    "tags",
]

# =============================================
# Follows is Instagram graph fields to get data
# =============================================

IG_BUSINESS_USER_PUBLIC_FIELDS = [
    "id",
    "biography",
    "name",
    "username",
    "profile_picture_url",
    "followers_count",
    "media_count",
]

IG_BUSINESS_MEDIA_CHILDREN_PUBLIC_FIELDS = [
    "id",
    "media_type",
    "media_url",
    "permalink",
    "shortcode",
    "thumbnail_url",
    "timestamp",
]

IG_BUSINESS_MEDIA_PUBLIC_FIELDS = [
    "id",
    "caption",
    "comments_count",
    f"children{{{','.join(IG_BUSINESS_MEDIA_CHILDREN_PUBLIC_FIELDS)}}}",
    "like_count",
    "media_type",
    "media_url",
    "permalink",
    "timestamp",
]

IG_BUSINESS_COMMENT_PUBLIC_FIELDS = [
    "id",
    "like_count",
    "text",
    "timestamp",
    "username",
]

IG_BUSINESS_REPLY_PUBLIC_FIELDS = IG_BUSINESS_COMMENT_PUBLIC_FIELDS

IG_BUSINESS_HASHTAG_PUBLIC_FIELDS = ["id", "name"]

IG_BUSINESS_CONTAINER_PUBLIC_FIELDS = ["id", "status_code"]

IG_BUSINESS_CONTENT_PUBLISH_LIMIT_FIELDS = ["config", "quota_usage"]

IG_BUSINESS_MENTION_MEDIA_FIELDS = [
    "caption",
    "comments_count",
    "id",
    "like_count",
    "media_type",
    "media_url",
    "timestamp",
]

IG_BASIC_USER_FIELDS = [
    "account_type",
    "id",
    "media_count",
    "username",
]

IG_BASIC_MEDIA_CHILDREN_FIELDS = ["id", "media_type", "media_url"]

IG_BASIC_MEDIA_FIELDS = [
    "caption",
    "id",
    "media_type",
    "media_url",
    "permalink",
    "thumbnail_url",
    "timestamp",
    "username",
]



"""
    Apis for group.
"""

from typing import Dict, Optional, Union

import pyfacebook.utils.constant as const
from pyfacebook.api.base_resource import BaseResource
from pyfacebook.api.facebook.common_edges import (
    FeedEdge,
    AlbumsEdge,
    PhotosEdge,
    VideosEdge,
    LiveVideosEdge,
)
from pyfacebook.models.group import Group
from pyfacebook.utils.params_utils import enf_comma_separated


class FacebookGroup(
    BaseResource, FeedEdge, AlbumsEdge, PhotosEdge, VideosEdge, LiveVideosEdge
):
    def get_info(
        self,
        group_id: str,
        fields: Optional[Union[str, list, tuple]] = None,
        return_json=False,
    ) -> Union[Group, dict]:
        """
        Get information about a Facebook Group.

        :param group_id: ID for the group.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for page.
            Or return json data. Default is false.
        :return: Group information.
        """
        if fields is None:
            fields = const.GROUP_PUBLIC_FIELDS
        data = self.client.get_object(
            object_id=group_id, fields=enf_comma_separated(field="fields", value=fields)
        )
        if return_json:
            return data
        else:
            return Group.new_from_json_dict(data=data)

    def get_batch(
        self,
        ids: Union[str, list, tuple],
        fields: Optional[Union[str, list, tuple]] = None,
        return_json: bool = False,
    ) -> Union[Dict[str, Group], dict]:
        """
        Get batch groups information by ids.

        :param ids: IDs for the groups.
        :param fields: Comma-separated id string for data fields which you want.
            You can also pass this with an id list, tuple.
        :param return_json: Set to false will return a dataclass for page.
            Or return json data. Default is false.
        :return: Groups information.
        """

        ids = enf_comma_separated(field="ids", value=ids)

        if fields is None:
            fields = const.GROUP_PUBLIC_FIELDS

        data = self.client.get_objects(
            ids=ids, fields=enf_comma_separated(field="fields", value=fields)
        )
        if return_json:
            return data
        else:
            return {
                gp_id: Group.new_from_json_dict(item) for gp_id, item in data.items()
            }



"""
    Tests for comment.
"""

import responses


def test_comment_get_info(helpers, api):
    comment_id = "17892250648466172"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{comment_id}",
            json=helpers.load_json(
                "testdata/instagram/apidata/comments/comment_fields.json"
            ),
        )

        comment = api.comment.get_info(comment_id=comment_id)
        assert comment.id == comment_id

        comment_json = api.comment.get_info(
            comment_id=comment_id,
            fields="id,like_count,text,timestamp,username",
            return_json=True,
        )
        assert comment_json["id"] == comment_id


def test_comment_get_batch(helpers, api):
    comment_ids = ["17892250648466172", "17858154961981086"]

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}",
            json=helpers.load_json(
                "testdata/instagram/apidata/comments/comments_fields.json"
            ),
        )

        comments = api.comment.get_batch(ids=comment_ids)
        assert comments[comment_ids[0]].id == comment_ids[0]

        comments_json = api.comment.get_batch(
            ids=comment_ids,
            fields="id,like_count,text,timestamp,username",
            return_json=True,
        )
        assert comments_json[comment_ids[0]]["id"] == comment_ids[0]


def test_comment_get_replies(helpers, api):
    comment_id = "17874875428706419"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{comment_id}/replies",
            json=helpers.load_json("testdata/instagram/apidata/comments/replies.json"),
        )

        replies = api.comment.get_replies(comment_id=comment_id)
        assert len(replies.data) == 1

        replies_json = api.comment.get_replies(comment_id=comment_id, return_json=True)
        assert replies_json["data"][0]["id"] == "17847314687296283"


def test_reply_get_info(helpers, api):
    reply_id = "17892250648466172"

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}/{reply_id}",
            json=helpers.load_json(
                "testdata/instagram/apidata/comments/comment_fields.json"
            ),
        )

        reply = api.reply.get_info(reply_id=reply_id)
        assert reply.id == reply_id

        reply_json = api.reply.get_info(
            reply_id=reply_id,
            fields="id,like_count,text,timestamp,username",
            return_json=True,
        )
        assert reply_json["id"] == reply_id


def test_reply_get_batch(helpers, api):
    reply_ids = ["17892250648466172", "17858154961981086"]

    with responses.RequestsMock() as m:
        m.add(
            method=responses.GET,
            url=f"https://graph.facebook.com/{api.version}",
            json=helpers.load_json(
                "testdata/instagram/apidata/comments/comments_fields.json"
            ),
        )

        replies = api.reply.get_batch(ids=reply_ids)
        assert replies[reply_ids[0]].id == reply_ids[0]

        replies_json = api.reply.get_batch(
            ids=reply_ids,
            fields="id,like_count,text,timestamp,username",
            return_json=True,
        )
        assert replies_json[reply_ids[0]]["id"] == reply_ids[0]



"""
    Extension models for graph api
"""

from dataclasses import dataclass
from typing import List, Optional

from pyfacebook.models.base import BaseModel, field
from pyfacebook.models.extensions import Paging


@dataclass
class ProfilePictureSource(BaseModel):
    """
    A class representing Profile Picture Source

    Refer: https://developers.facebook.com/docs/graph-api/reference/profile-picture-source/
    """

    cache_key: Optional[str] = field()
    height: Optional[int] = field()
    is_silhouette: Optional[bool] = field()
    url: Optional[str] = field(repr=True)
    width: Optional[int] = field()


@dataclass
class Picture(BaseModel):
    """
    A class representing the Picture connection.

    Refer: https://developers.facebook.com/docs/graph-api/reference/page/picture
    """

    data: Optional[ProfilePictureSource] = field(repr=True)
    paging: Optional[Paging] = field()


@dataclass
class CoverPhoto(BaseModel):
    """
    A class representing Cover Photo

    Refer: https://developers.facebook.com/docs/graph-api/reference/cover-photo/
    """

    id: Optional[str] = field(repr=True, compare=True)
    offset_x: Optional[float] = field()
    offset_y: Optional[float] = field()
    source: Optional[str] = field(repr=True)


@dataclass
class ImageSource(BaseModel):
    """
    A class representing the Image Source.
    """

    height: Optional[int] = field()
    width: Optional[int] = field()
    src: Optional[str] = field(repr=True)
